"""Various tests for ensuring when we get provide a node_key we get the node.

Done by ensuring we can receive messages from the node.

Tests for function and method blocks, as well as placeholder specify methods.
"""
from typing import OrderedDict
import unittest

from deltalanguage.data_types import Void
from deltalanguage.runtime import DeltaRuntimeExit, DeltaPySimulator
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaGraph,
                                  DeltaMethodBlock,
                                  NodeTemplate,
                                  placeholder_node_factory,
                                  PythonNode)


class Foo:

    def __init__(self):
        self.x = None

    @DeltaMethodBlock(node_key="node")
    def add_set_x(self, a: int, b: int, node: PythonNode = None) -> Void:
        self.x = node.receive('a') + node.receive('b')
        raise DeltaRuntimeExit


class TestNodeKey(unittest.TestCase):
    def test_func(self):
        @DeltaBlock(node_key="node")
        def add_assert(a: int, b: int, node: PythonNode = None) -> Void:
            self.assertEqual(node.receive('a') + node.receive('b'), 9)
            raise DeltaRuntimeExit
        with DeltaGraph() as test_graph:
            add_assert(a=4, b=5)
        rt = DeltaPySimulator(test_graph)
        rt.run()

    def test_method(self):
        foo = Foo()
        with DeltaGraph() as test_graph:
            foo.add_set_x(a=4, b=5)
        rt = DeltaPySimulator(test_graph)
        rt.run()
        self.assertEqual(foo.x, 9)

    def test_placeholder_func(self):
        @DeltaBlock(node_key="node")
        def add_assert(a: int, b: int, node: PythonNode = None) -> Void:
            self.assertEqual(node.receive('a') + node.receive('b'), 9)
            raise DeltaRuntimeExit
        with DeltaGraph() as test_graph:
            n = placeholder_node_factory(a=4, b=5)
        n.specify_by_func(add_assert, node_key="node")
        rt = DeltaPySimulator(test_graph)
        rt.run()

    def test_placeholder_method(self):
        foo = Foo()
        with DeltaGraph() as test_graph:
            n = placeholder_node_factory(a=4, b=5)
        n.specify_by_method(Foo.add_set_x, foo, node_key="node")
        rt = DeltaPySimulator(test_graph)
        rt.run()
        self.assertEqual(foo.x, 9)


test_template1 = NodeTemplate(name="NodeKeyTestTemplate",
                              in_params=OrderedDict([('a', int), ('b', int)]),
                              out_type=Void, node_key='node')


class Foo_T:

    def __init__(self):
        self.x = None

    @DeltaMethodBlock(node_key="node")
    def add_set_x_t(self, a: int, b: int, node: PythonNode = None) -> Void:
        self.x = node.receive('a') + node.receive('b')
        raise DeltaRuntimeExit


@DeltaBlock(template=test_template1, node_key="node")
def add_print_t(a: int, b: int, node: PythonNode = None) -> Void:
    print(node.receive('a') + node.receive('b'))
    raise DeltaRuntimeExit


class TestNodeKeyTemplated(unittest.TestCase):
    """Test that node key specification can occur correctly
    when a NodeTemplate is used.
    """

    def test_template_different_selected(self):
        """Test that node key is still correctly used on a multi-body
        node that has its body selected different from the one that was 
        specified via the constructor.
        """
        foo = Foo_T()
        with DeltaGraph() as test_graph:
            n1 = add_print_t(a=4, b=5)

        n1.add_body(foo.add_set_x_t)
        test_graph.select_bodies(preferred=["add_set_x_t"])
        DeltaPySimulator(test_graph).run()
        self.assertIn("add_set_x_t", n1.body.access_tags)
        self.assertEqual(foo.x, 9)


if __name__ == "__main__":
    unittest.main()



"""These tests are to demonstrate how some expected behaviour between
an interactive node and a migen node with a FSM may fail due to a signal
being unavailable on specific clock cycles.
"""

import unittest

from migen import FSM, If, NextState, NextValue

from deltalanguage.data_types import (DInt,
                                      DOptional,
                                      make_forked_return)
from deltalanguage.lib import StateSaver
from deltalanguage.runtime import DeltaPySimulator
from deltalanguage.wiring import (DeltaGraph,
                                  Interactive,
                                  MigenNodeTemplate,
                                  PythonNode)


TestBenchForkT, TestBenchFork = make_forked_return(
    {
        'inp': int,
        'trigger': int
    }
)


@Interactive({}, TestBenchForkT)
def test_bench_no_trigger(node: PythonNode):
    """This node sends 2 consecutive messages, the first is a data signal, the
    second is a trigger to move out of a specific state of the migen FSM.
    """

    node.send(TestBenchFork(15, 0))
    node.send(TestBenchFork(0, 1))


@Interactive({}, TestBenchForkT)
def test_bench_yes_trigger(node: PythonNode):
    """This node sends 3 consecutive messages, the first is a data signal, the
    second is the same data signal, and the third is a trigger to move out of
    a specific state of the migen FSM.
    """

    node.send(TestBenchFork(15, 0))
    # make inp.data available for an extra cycle of migen clock
    node.send(TestBenchFork(15, 0))
    node.send(TestBenchFork(0, 1))


class TestMigenNode(MigenNodeTemplate):
    """A migen node that accepts a data signal, and a trigger signal that
    moves the FSM out of the "LOADING" state.
    """

    def migen_body(self, template):

        inp = template.add_pa_in_port('inp', DOptional(DInt()))
        trigger = template.add_pa_in_port('trigger', DOptional(DInt()))
        out = template.add_pa_out_port('out', DInt())

        # Declare input and output ports always happy to receive/transmit data
        self.comb += (
            inp.ready.eq(1),
            trigger.ready.eq(1),
            out.ready.eq(1),
        )

        commander_fsm = FSM(reset_state="IDLE")
        self.submodules.commander_fsm = commander_fsm

        commander_fsm.act(
            "IDLE",
            If(
                inp.valid == 1,
                NextState("LOADING")
            )
        )

        commander_fsm.act(
            "LOADING",
            If(
                trigger.valid & trigger.data == 1,
                NextState("RETURN")
            ).Else(
                NextValue(out.data, out.data + inp.data),
            )
        )

        commander_fsm.act(
            "RETURN",
            NextValue(out.valid, 1),
            NextState("IDLE")
        )


class MigenInteractiveCommsTest(unittest.TestCase):

    def setUp(self):

        self.saver = StateSaver(int)

    def test_migen_trigger_fails(self):
        """Assert that when the `test_bench_no_trigger` node sends data, the
        ouput of the migen node is 0 due to the data not loading into output
        signal properly.
        """

        with DeltaGraph() as graph:

            test_bench_output = test_bench_no_trigger.call()

            c1 = TestMigenNode(
                tb_num_iter=10
            ).call(
                inp=test_bench_output.inp,
                trigger=test_bench_output.trigger
            )

            self.saver.save_and_exit(c1.out)

        rt = DeltaPySimulator(graph)
        rt.run()

        self.assertEqual(self.saver.saved[0], 0)

    def test_migen_trigger_succeeds(self):
        """Assert that when the `test_bench_ye_trigger` node sends data, the
        ouput of the migen node is 15 since the data signal is available on the
        particular clock cycle that loads in the data.
        """

        with DeltaGraph() as graph:

            test_bench_output = test_bench_yes_trigger.call()

            c1 = TestMigenNode(
                tb_num_iter=10
            ).call(
                inp=test_bench_output.inp,
                trigger=test_bench_output.trigger
            )

            self.saver.save_and_exit(c1.out)

        rt = DeltaPySimulator(graph)
        rt.run()

        self.assertEqual(self.saver.saved[0], 15)


if __name__ == "__main__":
    unittest.main()



"""Testing how nodes get turned to constant nodes, such as those with
a PyConstBody.

A node can be tagged with allow_contan==True, but it's not the only condition.
It's also important where this node takes the inputs from.
Below we consided these cases.

These tests do not test how the resulting graphs are executed, only
their construction.
"""

import unittest
from collections import OrderedDict

from test._utils import (return_12,
                         add,
                         add_non_const,
                         return_2)

from deltalanguage.wiring import (DeltaBlock,
                                  DeltaGraph,
                                  PyConstBody,
                                  PyFuncBody,
                                  NodeTemplate)


@DeltaBlock(allow_const=True)
def foo_const(a: int) -> int:
    return a


@DeltaBlock(allow_const=True)
def foo_const_2(a: int) -> int:
    return a


@DeltaBlock(allow_const=False)
def foo_non_const(a: int) -> int:
    return a


class ConstFolding(unittest.TestCase):
    """Test that single-bodies nodes are folded correctly during graph 
    construction.
    """

    def test_simple_folding_const_inputs(self):
        """Test that nodes producing constants are constant nodes."""

        with DeltaGraph() as graph:
            n1 = add(4, 3)
            n2 = add_non_const(4, 2)

        # nodes producing constants are turned to constant nodes
        for node in graph.find_node_by_name('node'):
            self.assertIsInstance(node.body, PyConstBody)

        # this one is turned to a constant node as well
        self.assertIsInstance(n1.body, PyConstBody)

        # this one would have done so as well, but we disallowed it
        self.assertIsInstance(n2.body, PyFuncBody)

        self.assertTrue(graph.check())

    def test_simple_folding_non_const_inputs(self):
        """Test that nodes with non constant inputs cannot fold to become
        constant.
        """

        with DeltaGraph() as graph:
            n1 = add(4, foo_non_const(3))
            n2 = add_non_const(4, foo_non_const(2))

        # nodes producing constants are turned to constant nodes
        for node in graph.find_node_by_name('node'):
            self.assertIsInstance(node.body, PyConstBody)

        # this one is a non-constant one now
        self.assertIsInstance(n1.body, PyFuncBody)

        # this one is still a non-constant
        self.assertIsInstance(n2.body, PyFuncBody)

        self.assertTrue(graph.check())

    def test_simple_folding_chain(self):
        """Test that a chain of node tagged as constant will turn to
        constant nodes.
        """
        with DeltaGraph() as graph:
            n0 = return_2()
            n1 = foo_const(n0)
            n2 = foo_const(n1)

        self.assertIsInstance(n0.body, PyConstBody)
        self.assertIsInstance(n1.body, PyConstBody)
        self.assertIsInstance(n2.body, PyConstBody)

        self.assertTrue(graph.check())

    def test_simple_folding_chain_splitting(self):
        """Test that if an output of a constant node can be split between
        a constant and non-constan nodes, i.e. the source stays constant and
        only one of the destinations is constant.
        """
        with DeltaGraph() as graph:
            n0 = return_2()
            n1 = foo_const(n0)
            n2 = foo_const(n1)
            n3 = foo_non_const(n1)

        self.assertIsInstance(n0.body, PyConstBody)
        self.assertIsInstance(n1.body, PyConstBody)
        self.assertIsInstance(n2.body, PyConstBody)
        self.assertIsInstance(n3.body, PyFuncBody)

        self.assertTrue(graph.check())

    def test_forked_folding_all_const(self):
        """Check the case with a forked return.

        All the forked outputs go to const nodes ->
        source is a constant node as well.
        """
        with DeltaGraph() as graph:
            nums = return_12()
            p1 = foo_const(nums.x)
            p2 = foo_const(nums.y)

        self.assertIsInstance(nums.body, PyConstBody)
        self.assertIsInstance(p1.body, PyConstBody)
        self.assertIsInstance(p2.body, PyConstBody)

    def test_forked_folding_one_non_const(self):
        """Check the case with a forked return.

        One forked output goes to a non-const node ->
        the rest of destinations stay const.
        """
        with DeltaGraph() as graph:
            nums = return_12()
            p1 = foo_const(nums.x)
            p2 = foo_non_const(nums.y)

        self.assertIsInstance(nums.body, PyConstBody)
        self.assertIsInstance(p1.body, PyConstBody)
        self.assertIsInstance(p2.body, PyFuncBody)

        self.assertTrue(graph.check())

    def test_forked_folding_one_non_const_splitting(self):
        """Check the case with a forked return.

        Same as above but with splitting ->
        no funny business.
        """
        with DeltaGraph() as graph:
            nums = return_12()
            p1 = foo_const(nums.x)
            p2 = foo_non_const(nums.y)
            p3 = foo_const(nums.x)
            p4 = foo_const(nums.y)

        self.assertIsInstance(nums.body, PyConstBody)
        self.assertIsInstance(p1.body, PyConstBody)
        self.assertIsInstance(p2.body, PyFuncBody)
        self.assertIsInstance(p3.body, PyConstBody)
        self.assertIsInstance(p4.body, PyConstBody)


const_template = NodeTemplate(name="ConstFoldTestTemplate",
                              in_params=OrderedDict([('a', int)]),
                              out_type=int)


@DeltaBlock(template=const_template, allow_const=True)
def foo_const_1_t(a: int) -> int:
    return a


@DeltaBlock(template=const_template, allow_const=True)
def foo_const_2_t(a: int) -> int:
    return a


non_const_template = NodeTemplate(name="NonConstFoldTestTemplate",
                                  in_params=OrderedDict([('a', int)]),
                                  out_type=int)


@DeltaBlock(template=non_const_template, allow_const=False)
def foo_non_const_1_t(a: int) -> int:
    return a


@DeltaBlock(template=non_const_template, allow_const=False)
def foo_non_const_2_t(a: int) -> int:
    return a


semi_const_template = NodeTemplate(name="SemiConstFoldTestTemplate",
                                   in_params=OrderedDict([('a', int)]),
                                   out_type=int)


@DeltaBlock(template=semi_const_template, allow_const=False)
def foo_semi_const_1_t(a: int) -> int:
    return a


@DeltaBlock(template=semi_const_template, allow_const=True)
def foo_semi_const_2_t(a: int) -> int:
    return a


class ConstFoldingNodeTemplates(unittest.TestCase):
    """Test that nodes created via multi-bodies NodeTemplates are folded
    correctly during graph construction."""

    def test_simple_folding_chain(self):
        """Test that a chain of node tagged as constant will turn to
        constant nodes.
        """
        with DeltaGraph() as graph:
            n0 = return_2()
            n1 = foo_const_1_t(n0)
            n2 = foo_const_2_t(n1)

        self.assertIsInstance(n0.body, PyConstBody)
        self.assertIsInstance(n1.body, PyConstBody)
        self.assertIsInstance(n2.body, PyConstBody)

        self.assertTrue(graph.check())

    def test_simple_folding_non_const_inputs(self):
        """Test that nodes with non constant inputs cannot fold to become
        constant.
        """

        with DeltaGraph() as graph:
            n1 = add(4, foo_non_const_1_t(3))
            n2 = add_non_const(4, foo_non_const_1_t(2))

        # nodes producing constants are turned to constant nodes
        for node in graph.find_node_by_name('node'):
            self.assertIsInstance(node.body, PyConstBody)

        # this one is a non-constant one now
        self.assertIsInstance(n1.body, PyFuncBody)

        # this one is still a non-constant
        self.assertIsInstance(n2.body, PyFuncBody)

        self.assertTrue(graph.check())

    def test_cannot_fold_through_semi_const_multi_body(self):
        """If a node has both constant and non-constant bodies, then
        this node cannot be folded.

        .. note::
          This behaviors might change so only the selected body
          (during execution) will define if the node can/cannot be
          folded.
        """

        with DeltaGraph() as graph:
            n1 = add(4, foo_semi_const_1_t(3))
            n2 = add_non_const(4, foo_semi_const_1_t(2))

        # nodes producing constants are turned to constant nodes
        for node in graph.find_node_by_name('node'):
            self.assertIsInstance(node.body, PyConstBody)

        # this one is a non-constant one now
        self.assertIsInstance(n1.body, PyFuncBody)

        # this one is still a non-constant
        self.assertIsInstance(n2.body, PyFuncBody)

        self.assertTrue(graph.check())


if __name__ == "__main__":
    unittest.main()



"""Special data types."""

from typing import (Any,
                    Dict,
                    Iterable,
                    List,
                    NamedTuple,
                    Tuple,
                    Type,
                    Union)

from deltalanguage._utils import NamespacedName
from ._exceptions import DeltaTypeError


class DSize:
    """Size of a Deltaflow type.

    Main usage is in :py:attr:`BaseDeltaType.size` and
    :py:attr:`DArray.length`.

    Parameters
    ----------
    size : Union[int, NamespacedName]
        It could be a definite size, or a
        placeholder for the actual size to be specified later.
    """

    def __init__(self, size: Union[int, NamespacedName]):
        self.is_placeholder = isinstance(size, NamespacedName)
        self.val = size

        if not self.is_placeholder and self.val < 0:
            raise ValueError("Only non-negative sizes for types are allowed.")

    def __repr__(self) -> str:
        if self.is_placeholder:
            return f"({self.val})"
        else:
            return f"{self.val}"

    def __eq__(self, other):
        if type(other) is DSize:
            return (self.is_placeholder == other.is_placeholder
                    and self.val == other.val)
        return False

    def __ge__(self, other):
        if type(other) is DSize:
            return (self.is_placeholder == other.is_placeholder
                    and self.val >= other.val)
        return False

    def __gt__(self, other):
        if self.is_placeholder or other.is_placeholder:
            raise ValueError("Please define DSize")
        return self.val > other.val

    def __hash__(self):
        return hash((self.is_placeholder, self.val))

    def __add__(self, other):
        if self.is_placeholder or other.is_placeholder:
            raise ValueError("Please define DSize")
        return DSize(self.val + other.val)

    def __iadd__(self, other):
        if self.is_placeholder or other.is_placeholder:
            raise ValueError("Please define DSize")
        self.val += other.val
        return self

    def __sub__(self, other):
        if self.is_placeholder or other.is_placeholder:
            raise ValueError("Please define DSize")
        return DSize(self.val - other.val)

    def __isub__(self, other):
        if self.is_placeholder or other.is_placeholder:
            raise ValueError("Please define DSize")

        self.val -= other.val

        if self.val < 0:
            raise ValueError("Only non-negative sizes for types are allowed.")

        return self

    def __mul__(self, other):
        if self.is_placeholder:
            raise ValueError("Please define DSize")

        if isinstance(other, int):
            return DSize(self.val * other)
        elif type(other) is DSize:
            return DSize(self.val * other.val)
        else:
            raise DeltaTypeError("Unsupported data type")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __imul__(self, other):
        if self.is_placeholder:
            raise ValueError("Please define DSize")

        if isinstance(other, int):
            self.val *= other
        elif type(other) is DSize:
            self.val *= other.val
        else:
            raise DeltaTypeError("Unsupported data type")

        return self

    @property
    def in_bytes(self) -> int:
        if self.val % 8 == 0:
            return self.val//8
        else:
            raise ValueError(f'{self} bits need a fractional number of bytes')


class Void:
    """Used in cases if a node does not have and output channel.

    Examples
    --------
    In this case there is no need to even add a ``return`` statement.

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.DeltaBlock()
        ... def foo(a: int) -> dl.Void:
        ...     print('Hello Quantum World!')

    If the node may or may not send a message
    you need to create a channel by specifying the output type:

    .. code-block:: python

        >>> @dl.DeltaBlock()
        ... def bar(a: int) -> int:
        ...     if a%2 == 0:
        ...         return a//2
        ...     else:
        ...         pass
    """
    pass



import unittest
from collections import OrderedDict

from deltalanguage.lib import StateSaver
from deltalanguage.runtime import DeltaPySimulator, DeltaRuntimeExit
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaGraph,
                                  DeltaMethodBlock,
                                  MigenNodeTemplate,
                                  NodeTemplate,
                                  Interactive,
                                  RealNode)
from deltalanguage.data_types import DOptional
from deltalanguage.wiring._node_classes.node_bodies import PyMigenBody


test_template1 = NodeTemplate(name="AdderTemplate",
                              in_params=OrderedDict([('a', int), ('b', int)]),
                              out_type=int)


@DeltaBlock(test_template1, allow_const=False)
def simple_add(a: int, b: int) -> int:
    return a + b


@DeltaBlock(test_template1, allow_const=False)
def over_complex_add(a: int, b: int) -> int:
    return 2*a + 2*b - a - b


@Interactive(template=test_template1,
             in_params=OrderedDict([('a', int), ('b', int)]),
             out_type=int)
def broken_adder(node: RealNode):
    a = node.receive('a')
    b = node.receive('b')
    node.send(a+b+1)


class OpCacher():

    def __init__(self):
        self._add_cache = {}

    @DeltaMethodBlock(test_template1)
    def cached_add(self, a: int, b: int) -> int:
        if (a, b) in self._add_cache:
            return self._add_cache[(a, b)]
        else:
            result = a+b
            self._add_cache[(a, b)] = result
            return result


class NodeTempalteViaDecoratorsTest(unittest.TestCase):
    """Tests for the NodeTemplate and its use in associating multiple
    constructors/bodies together.
    """

    def test_via_DeltaBlock(self):
        saver = StateSaver()
        with DeltaGraph() as graph:
            n1 = simple_add(4, 3)
            saver.save_and_exit(n1)
        self.assertEqual(len(n1.bodies), 3)
        self.assertIn("simple_add", n1.body.access_tags)
        DeltaPySimulator(graph).run()
        self.assertEqual([7], saver.saved)

    def test_via_DeltaMethodBlock(self):
        saver = StateSaver()
        cacher = OpCacher()
        with DeltaGraph() as graph:
            n1 = cacher.cached_add(1, 3)
            saver.save_and_exit(n1)
        self.assertEqual(len(n1.bodies), 4)
        self.assertIn("cached_add", n1.body.access_tags)
        DeltaPySimulator(graph).run()
        self.assertEqual([4], saver.saved)

    def test_via_Interactive(self):
        saver = StateSaver()
        with DeltaGraph() as graph:
            n1 = broken_adder.call(a=1, b=3)
            saver.save_and_exit(n1)
        self.assertEqual(len(n1.bodies), 3)
        self.assertIn("broken_adder", n1.body.access_tags)
        DeltaPySimulator(graph).run()
        self.assertEqual([5], saver.saved)


m_template = NodeTemplate(name="MTemplate",
                          in_params=OrderedDict([('a', int), ('b', int)]),
                          out_type=int)


class AMigenNode(MigenNodeTemplate):
    def migen_body(self, template):
        template.add_pa_in_port('a', DOptional(int))
        template.add_pa_in_port('b', DOptional(int))


class NodeTemplateViaMigenClass(unittest.TestCase):
    """Test that migen body templates can be added to node templates
    using the migen node template constructor.
    """

    def test_via_migen_init(self):

        m_template = NodeTemplate(
            name="MTemplate",
            in_params=OrderedDict([('a', DOptional(int)),
                                   ('b', DOptional(int))])
        )

        @DeltaBlock(m_template, allow_const=False)
        def m_simple_add(a: DOptional(int), b: DOptional(int)):
            raise DeltaRuntimeExit

        m_maker = AMigenNode(node_template=m_template)
        with DeltaGraph():
            n1 = m_maker.call(a=1, b=3)
        self.assertEqual(len(n1.bodies), 2)
        self.assertIn(PyMigenBody, n1.body.access_tags)


class NodeTemplateTest(unittest.TestCase):
    """Some core behaviours of node templates tested
    """

    def test_multiple_nodes_same_template(self):
        """Test that the same NodeTemplate can be used to create
        multiple distinct nodes for the same graph.
        """
        saver = StateSaver()
        with DeltaGraph() as graph:
            n1 = over_complex_add(1, 3)
            n2 = simple_add(4, n1)
            saver.save_and_exit(n2)

        self.assertEqual(len(n1.bodies), 3)
        self.assertEqual(len(n2.bodies), 3)
        self.assertIn("over_complex_add", n1.body.access_tags)
        self.assertIn("simple_add", n2.body.access_tags)

        DeltaPySimulator(graph).run()
        self.assertEqual([8], saver.saved)

    def test_select_different_from_constructor(self):
        """Test that it is possible to select a different body than the
        one associated with the used constructor.
        """
        saver = StateSaver()
        with DeltaGraph() as graph:
            n1 = simple_add(1, 3)
            saver.save_and_exit(n1)

        graph.select_bodies(preferred=["broken_adder"])
        self.assertEqual(len(n1.bodies), 3)
        self.assertIn("broken_adder", n1.body.access_tags)

        DeltaPySimulator(graph).run()
        self.assertEqual([5], saver.saved)


class NodeTemplateDefaultCall(unittest.TestCase):
    """Tests for when the default call is used to create nodes with one or
    more bodies.

    .. note::
      See `test_no_body_node.py` for when default call is used to create nodes
      with no bodies.
    """

    def test_via_default_call(self):
        saver = StateSaver()
        with DeltaGraph() as graph:
            n1 = test_template1.call(9, 2)
            saver.save_and_exit(n1)
        self.assertEqual(len(n1.bodies), 3)
        self.assertIn("simple_add", n1.body.access_tags)
        DeltaPySimulator(graph).run()
        self.assertEqual([11], saver.saved)


class OpCacher2():

    def __init__(self):
        self._add_cache = {}

    @DeltaMethodBlock()
    def cached_add_2(self, a: int, b: int) -> int:
        if (a, b) in self._add_cache:
            return self._add_cache[(a, b)]
        else:
            result = a+b
            self._add_cache[(a, b)] = result
            return result


class AMigenNode2(MigenNodeTemplate):
    def migen_body(self, template):
        template.add_pa_in_port('a', DOptional(int))
        template.add_pa_in_port('b', DOptional(int))


class NodeTemplateAddConstructor(unittest.TestCase):
    """Tests for ``add_constructor`` method for adding new constructors
    to a ``NodeTemplate`` without re-defining the original constructor.
    """

    def test_add_func(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(allow_const=False)
        def simple_add_2(a: int, b: int) -> int:
            return a + b

        test_template1.add_constructor(simple_add_2)

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 1)
        self.assertIn('simple_add_2', n1.body.access_tags)

    def test_add_method(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(template=test_template1, allow_const=False)
        def simple_add_2(a: int, b: int) -> int:
            return a + b

        test_template1.add_constructor(OpCacher2.cached_add_2)

        with DeltaGraph():
            n1 = OpCacher2().cached_add_2(2, 3)

        self.assertEqual(len(n1.bodies), 2)
        self.assertIn('cached_add_2', n1.body.access_tags)

    def test_add_interactive(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @Interactive(in_params=OrderedDict([('a', int), ('b', int)]),
                     out_type=int)
        def broken_adder_2(node: RealNode):
            a = node.receive('a')
            b = node.receive('b')
            node.send(a+b+1)

        test_template1.add_constructor(broken_adder_2)

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 1)
        self.assertIn('broken_adder_2', n1.body.access_tags)

    def test_add_migen(self):
        test_template1 = NodeTemplate(
            name="test_1",
            in_params=OrderedDict([('a', DOptional(int)),
                                   ('b', DOptional(int))])
        )

        test_template1.add_constructor(AMigenNode2())

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 1)
        self.assertIn(PyMigenBody, n1.body.access_tags)

    def test_add_with_existing_other_template(self):
        """Test for when the constructor we are adding is already associated
        with some other ``NodeTemplate``.
        """
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        test_template2 = NodeTemplate(name="test_2",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(test_template2, allow_const=False)
        def simple_add_2(a: int, b: int) -> int:
            return a + b

        test_template1.add_constructor(simple_add_2)

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 1)
        self.assertIn('simple_add_2', n1.body.access_tags)

    def test_add_with_existing_same_template(self):
        """Test for when the constructor we are adding is already on the
        ``NodeTemplate``. The constructor should not be added twice.
        """
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(test_template1, allow_const=False)
        def simple_add_2(a: int, b: int) -> int:
            return a + b

        test_template1.add_constructor(simple_add_2)

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 1)
        self.assertIn('simple_add_2', n1.body.access_tags)

    def test_add_invalid_constructor(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(allow_const=False)
        def simple_add_to_bool(a: int, b: int) -> bool:
            return bool(a + b)

        with self.assertRaises(ValueError):
            test_template1.add_constructor(simple_add_to_bool)


class InvalidNodeTemplate(unittest.TestCase):
    """Test that errors are raised when node templates
    are created in some way that is invalid.
    """

    def test_in_params_enforced(self):
        """Test to ensure in params must match when associating
        constructors with a NodeTemplate
        """
        test_template2 = NodeTemplate(name="test",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        with self.assertRaises(ValueError):
            @DeltaBlock(test_template2, allow_const=False)
            def _test1(a: int, b: bool) -> int:
                return a + b

        with self.assertRaises(ValueError):
            @DeltaBlock(test_template2, allow_const=False)
            def _test2(wrong_name: int, b: int) -> int:
                return wrong_name + b

    def test_in_params_enforced_optional(self):
        """Test to ensure in params must match when associating
        constructors with a NodeTemplate
        """
        test_template2 = NodeTemplate(
            name="test",
            in_params=OrderedDict(
                [('a', DOptional(int)), ('b', DOptional(int))]),
            out_type=int
        )

        with self.assertRaises(ValueError):
            @DeltaBlock(test_template2, allow_const=False)
            def _test1(a: DOptional(int), b: int) -> int:
                return a + b

        with self.assertRaises(ValueError):
            @ DeltaBlock(test_template2, allow_const=False)
            def _test2(wrong_name: int, b: int) -> int:
                return wrong_name + b

    def test_out_type_enforced(self):
        """Test to ensure out type must match when associating
        constructors with a NodeTemplate
        """
        test_template2 = NodeTemplate(name="test",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        with self.assertRaises(ValueError):
            @DeltaBlock(test_template2, allow_const=False)
            def _test1(a: int, b: int) -> bool:
                return a + b


class NodeTemplateMerge(unittest.TestCase):
    """Test that node templates can be merged correctly or raise errors when
    merge is not appropriate.
    """

    def __init__(self, methodName):
        super().__init__(methodName)
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(test_template1, allow_const=False)
        def simple_add(a: int, b: int) -> int:
            return a + b

        self.template_1_body = simple_add.template
        self.test_template1 = test_template1

    def test_merge_adds_bodies(self):
        """Test to ensure in params must match when associating
        constructors with a NodeTemplate.
        """
        test_template3 = NodeTemplate(name="test_3",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(test_template3, allow_const=False)
        def simple_add(a: int, b: int) -> int:
            return a + b

        test_template2 = NodeTemplate(name="test_2",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(test_template2, allow_const=False)
        def simple_add_2(a: int, b: int) -> int:
            return a + b

        test_template3.merge(test_template2)
        self.assertEqual(len(test_template3._body_templates), 2)
        self.assertIn(simple_add.template, test_template3._body_templates)
        self.assertIn(simple_add_2.template, test_template3._body_templates)

    def test_merge_use_in_graph(self):
        """Test to ensure in params must match when associating
        constructors with a NodeTemplate.
        """
        test_template3 = NodeTemplate(name="test_3",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(test_template3, allow_const=False)
        def simple_add(a: int, b: int) -> int:
            return a + b

        test_template2 = NodeTemplate(name="test_2",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(test_template2, allow_const=False)
        def simple_add_2(a: int, b: int) -> int:
            return a + b

        test_template3.merge(test_template2)

        saver = StateSaver()
        with DeltaGraph() as graph:
            n1 = simple_add_2(4, 73)
            saver.save_and_exit(n1)

        self.assertEqual(len(n1.bodies), 2)

        graph.select_bodies(preferred=['simple_add'])
        self.assertIn("simple_add", n1.body.access_tags)
        DeltaPySimulator(graph).run()
        self.assertEqual([77], saver.saved)

    def test_in_params_enforced_merge(self):
        """Test to ensure in params must match when associating
        constructors with a NodeTemplate.
        """
        test_template2 = NodeTemplate(name="test_2",
                                      in_params=OrderedDict(
                                          [('a', int), ('c', int)]),
                                      out_type=int)

        with self.assertRaises(ValueError):
            self.test_template1.merge(test_template2)

        with self.assertRaises(ValueError):
            test_template2.merge(self.test_template1)

    def test_out_type_enforced_merge(self):
        """Test to ensure in params must match when associating
        constructors with a NodeTemplate.
        """
        test_template2 = NodeTemplate(name="test_2",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=bool)

        with self.assertRaises(ValueError):
            self.test_template1.merge(test_template2)

        with self.assertRaises(ValueError):
            test_template2.merge(self.test_template1)

    def test_self_merge(self):
        """Test to ensure merge self does not result in odd behaviour."""
        self.test_template1.merge(self.test_template1)
        self.assertEqual(len(self.test_template1._body_templates), 1)
        self.assertIn(self.template_1_body,
                      self.test_template1._body_templates)


if __name__ == "__main__":
    unittest.main()



"""Port classes for bound ports between nodes.
Each node stores references to all of its in and out ports.
"""

from typing import NamedTuple, Union

import dill

from deltalanguage._utils import NamespacedName
from deltalanguage.data_types import BaseDeltaType, DOptional

from .abstract_node import AbstractNode


class InPort(
    NamedTuple(
        "InPort",
        [
            ("port_name", NamespacedName),
            ("port_type_", Union[BaseDeltaType, DOptional]),
            ("node", AbstractNode),
            ("in_port_size", int)
        ]
    )
):
    """Input port class.

    An instance of this class is created for every argument (input) of a node
    in :py:class:`DeltaGraph`.

    Attributes
    ----------
    port_name
        The name of the argument this port is supplying.
    port_type_
        The expected type of the value to be supplied.

        PRIVATE! Please use ``port_type`` and ``is_optional``.
    node
        The node this port is supplying to.
    in_port_size
        Limit for how many inputs this port can have enqueued.
        If a node tries to push more to this port's queue then it is blocked.


    .. TODO::
        ``InPort`` and ``OutPort`` can be implemented without ``NamedTuple``,
        thus ``port_type_`` hack won't be needed.
    """

    def __str__(self):
        ret = f"{self.port_name.n_index}, {self.port_type}"
        if self.is_optional:
            ret += ", optional"

        return ret

    def __repr__(self):
        return f"({self.port_name}, {self.port_type}, " +\
            f"{self.in_port_size}, {self.is_optional})"

    @property
    def index(self):
        """Index of this port."""
        return self.port_name.n_index

    @property
    def node_name(self):
        """Node name of this port."""
        return self.port_name.domain

    @property
    def port_type(self):
        if isinstance(self.port_type_, DOptional):
            return self.port_type_.type
        else:
            return self.port_type_

    @property
    def is_optional(self):
        """Returns ``True`` if this port is created for an optional input."""
        return isinstance(self.port_type_, DOptional)

    def capnp(self, capnp_in_port):
        """Serialise this port in ``capnp`` structure.

        Parameters
        ----------
        capnp_in_port
            The ``capnp`` object of this in port.
        """
        capnp_in_port.name = self.port_name.n_index
        capnp_in_port.type = dill.dumps(self.port_type)
        capnp_in_port.optional = self.is_optional

    def capnp_wiring(self, nodes, capnp_wire):
        """Serialise the wire that this port connects to.

        Parameters
        ----------
        nodes
            List of nodes in capnp structure.
            Used to find required indexes.
        capnp_wire
            The capnp object of this wire.
        """
        for i, node in enumerate(nodes):
            if node.name == self.port_name.domain:
                capnp_wire.destNode = i
                for j, in_port in enumerate(node.inPorts):
                    if in_port.name == self.port_name.n_index:
                        capnp_wire.destInPort = j
                        break
                break
        capnp_wire.direct = (self.in_port_size > 0)


class OutPort(
    NamedTuple(
        "OutPort",
        [
            ("port_name", NamespacedName),
            ("port_type_", BaseDeltaType),
            ("destination", InPort),
            ("node", AbstractNode)
        ]
    )
):
    """Output port class.

    An instance of this class is created for every return (output) of a node
    in :py:class:`DeltaGraph`.

    Attributes
    ----------
    port_name
        The name of the part of the node output this port will export.
    port_type_
        The type of the part of the node output this port will export.
        This could be the type of the whole output.

        PRIVATE! Please use ``port_type``.
    destination
        Reference to the port that this node is exporting to.
    node
        The node this port is exporting from
    """

    def __str__(self):
        name = self.port_name.n_index if self.port_name.n_index is not None \
            else "output"

        ret = f"{name}, {self.port_type} -> {self.destination.port_name}"
        return ret

    def __repr__(self):
        return f"({self.destination.port_name}, {self.port_name}, " +\
            f"{self.port_type})"

    @property
    def port_type(self):
        if isinstance(self.port_type_, DOptional):
            raise TypeError('OutPort cannot be optional')
        else:
            return self.port_type_

    @property
    def index(self):
        """Index of this port."""
        return self.port_name.n_index

    @property
    def dest_port_index(self):
        """Index of destination port."""
        return self.destination.index

    @property
    def dest_port_name(self):
        """Name of destination port (node name + index)."""
        return self.destination.port_name

    @property
    def dest_node_name(self):
        """Name of destination node."""
        return self.destination.node_name

    def capnp(self, capnp_out_port):
        """Serialise this port in ``capnp`` structure.

        Parameters
        ----------
        capnp_in_port
            The ``capnp`` object of this in port.
        """
        if self.port_name.n_index:
            capnp_out_port.name = self.port_name.n_index
        else:
            capnp_out_port.name = ""
        capnp_out_port.type = dill.dumps(self.port_type)

    def capnp_wiring(self, nodes, capnp_wire):
        """Serialise the wire that this port connects to.

        Parameters
        ----------
        nodes
            List of nodes in ``capnp`` structure.
            Used to find required indexes.
        capnp_wire
            The ``capnp`` object of this wire.
        """
        self.destination.capnp_wiring(nodes, capnp_wire)



"""Testing Void behaviour.
"""

import unittest

import deltalanguage as dl

from test._utils import printer


class VoidTest(unittest.TestCase):

    def test_error_thrown(self):
        """Nodes without output ports cannot produce output.

        .. note::
            This is very different from python's functions that return None
            by default.
        """
        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                no_message = printer(5)
                printer(no_message)


if __name__ == "__main__":
    unittest.main()



"""Module to describe the AbstractNode, ProxyNode and ForkedNode node classes.
"""

from abc import ABC, abstractmethod
from typing import Type

from .node_bodies import Body


class AbstractNode(ABC):
    """Abstract node class for :py:class:`DeltaGraph`.

    Used to derive :py:class:`ProxyNode` and :py:class:`RealNode`.

    ..
        Can't have more member data or methods at this level otherwise
        ProxyNodes would not work.
    """

    global_name_id = 0

    @classmethod
    def get_next_index(cls) -> int:
        """Fetch globally unique node names.

        Returns
        -------
        int
            a new node name that has not been used before
        """
        idx = AbstractNode.global_name_id
        AbstractNode.global_name_id += 1
        return idx

    @classmethod
    def reset_index(cls):
        """Reset the global node index.

        This is useful for tests where a fresh start is needed.
        """
        AbstractNode.global_name_id = 0

    @abstractmethod
    def add_out_port(self, port_destination, index=None):
        """Add out port.

        Parameters
        ----------
        port_destination : InPort
            The in-port that this out-port exports to.
        index : Optional[str]
            If this port is a fork then the type will be an indexed part
            of the output of the node.
        """
        pass

    @abstractmethod
    def add_in_port(self, port_name: str, in_type: Type):
        """Add in port.

        Parameters
        ----------
        port_name : str
            The name of the argument this port supplies.
        in_type : BaseDeltaType
            The type that is expected to be supplied for this port.
        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of this node as str.

        Returns
        -------
        str
        """
        pass

    @property
    @abstractmethod
    def body(self) -> Body:
        """The body this node represents.

        Returns
        -------
        Body
        """
        pass


class ProxyNode(AbstractNode):
    """Node class for a node that just refers to another node for all
    unspecified behaviours.

    Parameters
    ----------
    referee : AbstractNode
        Node that this proxy will act like. Defaults to ``None`` in which
        case any unhandled calls will throw exceptions.
    """

    def __init__(self, referee: AbstractNode = None):
        self.referee = referee

    def __getattr__(self, item):
        """Makes this node act like a proxy for the referee node.

        If this node defines some member behaviour it is fetched from here,
        otherwise it is fetched from the referee
        """
        return getattr(self.referee, item)

    def add_in_port(self, port_name: str, in_type: Type):
        """This function has been explicitly created as this is an abstract
        method calls add in-port on referee node.
        """
        self.referee.add_in_port(port_name, in_type)

    def add_out_port(self, port_destination, index=None):
        """This function has been explicitly created as this is an abstract
        method calls add out-port on referee node.
        """
        self.referee.add_out_port(port_destination, index)

    @property
    def name(self) -> str:
        return self.referee.name

    @property
    def body(self) -> Body:
        return self.referee.body


class ForkedNode(ProxyNode):
    """Node class to represent a pointer to a node that is forking its
    output to several other nodes.

    The main use of this class is to store the index at which this
    particular forked data can be found.

    Parameters
    ----------
    referee :
        The node that this node forks.
    index : str
        Index as string at which the data for this fork can be found
        on the referee nodes output.
    """

    def __init__(self, referee, index: str):
        super().__init__(referee)
        self.index = index

    def add_out_port(self, port_destination, index=None):
        """Add out port.

        Parameters
        ----------
        port_destination : InPort
            Where this out-port will go to, as InPort.
        index : Optional[str]
            Should always be none as we use this nodes stored index to create
            the port on the referee.
        """
        self.referee.add_out_port(port_destination, self.index)



import math
import subprocess
import unittest

import dill

from deltalanguage.data_types import Void
from deltalanguage.lib import StateSaver
from deltalanguage.runtime import DeltaPySimulator, DeltaRuntimeExit
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaGraph,
                                  DeltaMethodBlock,
                                  PyConstBody,
                                  PyFuncBody,
                                  PyMethodBody)

from test._utils import return_2


class Adder:

    def __init__(self, x: int):
        self._x = x

    @DeltaMethodBlock(name="add_x")
    def add_x(self, a: int) -> int:
        return a + self._x


class NodeSerialisationTest(unittest.TestCase):

    def setUp(self):

        @DeltaBlock(name="add_const")
        def add_const(a: int, b: int) -> int:
            return a + b

        @DeltaBlock(name="add_func", allow_const=False)
        def add_func(a: int, b: int) -> int:
            return a + b

        @DeltaBlock(name="gcd", allow_const=False)
        def gcd(a: int, b: int) -> int:
            return math.gcd(a, b)

        add_2 = Adder(2)

        with DeltaGraph() as graph:
            add_const(a=4, b=3)
            add_func(a=5, b=4)
            add_2.add_x(a=3)
            gcd(a=14, b=7)

        node_independent_const = graph.nodes[2]
        node_independent_func = graph.nodes[5]
        node_method = graph.nodes[7]
        node_env_func = graph.nodes[10]

        # const independent func
        self.serialized_independent_const =\
            node_independent_const.body.as_serialized
        self.deserialized_node_independent_const = dill.loads(
            self.serialized_independent_const
        )

        # non-const independent func with environment
        self.serialized_env_func =\
            node_env_func.body.as_serialized

        # non-const independent func
        self.serialized_independent_func =\
            node_independent_func.body.as_serialized
        self.deserialized_node_independent_func = dill.loads(
            self.serialized_independent_func
        )

        # method
        self.serialized_method = node_method.body.as_serialized
        self.deserialized_node_method = dill.loads(
            self.serialized_method
        )

    def test_deserialization(self):
        self.assertEqual(
            type(self.deserialized_node_independent_const),
            PyConstBody
        )
        self.assertEqual(
            type(self.deserialized_node_independent_func),
            PyFuncBody
        )
        self.assertEqual(
            type(self.deserialized_node_method),
            PyMethodBody
        )

    def test_load_new_environment_independent_const(self):

        # build up python script to run in separate environment
        python_string = "import dill\n"
        python_string += "deserialized_node_body = "
        python_string += f"dill.loads({self.serialized_independent_const})\n"
        python_string += "print(deserialized_node_body.eval())"

        # begin python process in separate environment
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(7, int(output))

    def test_load_new_environment_independent_nonconst(self):

        # build up python script to run in separate environment
        python_string = "import dill\n"
        python_string += "deserialized_node_body = "
        python_string += f"dill.loads({self.serialized_independent_func})\n"
        python_string += "print(deserialized_node_body.eval(5, 4))"

        # begin python process in separate environment
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(9, int(output))

    def test_load_new_environment_env_func(self):

        # build up python script to run in separate environment
        python_string = "import dill\n"
        python_string += "deserialized_node_body = "
        python_string += f"dill.loads({self.serialized_env_func})\n"
        python_string += "print(deserialized_node_body.eval(14, 21))"

        # begin python process in separate environment
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(7, int(output))

    def test_load_new_environment_method(self):

        # build up python script to run in separate environment
        python_string = "import dill\n"
        python_string += "deserialized_node_body = "
        python_string += f"dill.loads({self.serialized_method})\n"
        python_string += "print(deserialized_node_body.eval(3))"

        # begin python process in separate environment
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(5, int(output))

    def test_load_and_runtime_new_environment_independent_const(self):

        # build up python script to run in separate environment
        python_string = """
import dill

from deltalanguage.data_types import Void
from deltalanguage.runtime import DeltaPySimulator, DeltaRuntimeExit
from deltalanguage.wiring import DeltaBlock, DeltaGraph

@DeltaBlock(name="add")
def add(a: int, b: int) -> int:
    return a + b


@DeltaBlock(name="print_exit", allow_const=False)
def print_then_exit(to_print: object) -> Void:
    print(to_print)
    raise DeltaRuntimeExit


with DeltaGraph() as graph:
    added = add(a=4, b=3)
    print_then_exit(to_print=added)
"""

        python_string += f"""
deserialized_node = dill.loads({self.serialized_independent_const})

graph.nodes[2]._body = deserialized_node

rt = DeltaPySimulator(graph)
rt.run()

"""

        # begin python process in separate environment
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(7, int(output))

    def test_load_and_runtime_new_environment_env_func(self):

        # build up python script to run in separate environment
        python_string = """
import dill

from deltalanguage.data_types import Void
from deltalanguage.runtime import DeltaPySimulator, DeltaRuntimeExit
from deltalanguage.wiring import DeltaBlock, DeltaGraph

@DeltaBlock(name="add", allow_const=False)
def add(a: int, b: int) -> int:
    return a + b


@DeltaBlock(name="print_exit", allow_const=False)
def print_then_exit(to_print: object) -> Void:
    print(to_print)
    raise DeltaRuntimeExit


with DeltaGraph() as graph:
    added = add(a=14, b=21)
    print_then_exit(to_print=added)
"""

        python_string += f"""
deserialized_node = dill.loads({self.serialized_env_func})

graph.nodes[2]._body = deserialized_node

rt = DeltaPySimulator(graph)
rt.run()

"""

        # begin python process in separate environment
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(7, int(output))

    def test_load_and_runtime_new_environment_method(self):

        # build up python script to run in separate environment
        python_string = """
import dill

from deltalanguage.data_types import Void
from deltalanguage.runtime import DeltaPySimulator, DeltaRuntimeExit
from deltalanguage.wiring import DeltaBlock, DeltaMethodBlock, DeltaGraph

class Adder:

    def __init__(self, x: int):
        self._x = x

    @DeltaMethodBlock(name="add_x")
    def add_x(self, a: int) -> int:
        return a + self._x


@DeltaBlock(name="print_exit", allow_const=False)
def print_then_exit(to_print: object) -> Void:
    print(to_print)
    raise DeltaRuntimeExit

add_2 = Adder(2)

with DeltaGraph() as graph:
    added_2 = add_2.add_x(a=3)
    print_then_exit(to_print=added_2)
"""

        python_string += f"""
deserialized_node = dill.loads({self.serialized_method})

graph.nodes[1]._body = deserialized_node

rt = DeltaPySimulator(graph)
rt.run()

"""

        # begin python process in separate environment
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(5, int(output))


class StateSaverSerialisationTest(unittest.TestCase):

    def test_serialize_state_saver(self):
        """Serialise a state saver and load it in a new instance."""
        saver = StateSaver(t=int, verbose=True)
        with DeltaGraph() as test_graph:
            saver_node = saver.save(return_2())
        saver_body = saver_node.body.as_serialized
        python_string = f"""
import dill
saver_body = dill.loads({saver_body})
saver_body.eval(5)
"""
        p = subprocess.run(
            [r"python"],
            input=str.encode(python_string),
            stdout=subprocess.PIPE,
            check=False
        )
        output = p.stdout.decode()

        self.assertEqual(output, "saving 5\n")



if __name__ == "__main__":
    unittest.main()



import unittest

import deltalanguage as dl


class DecoratorNoInputTypeTest(unittest.TestCase):
    """If input type is not provided in node definition error is raised."""

    def test_DeltaBlock(self):
        with self.assertRaises(TypeError):
            @dl.DeltaBlock()
            def foo(a) -> int:
                return a

    def test_DeltaMethodBlock(self):
        with self.assertRaises(TypeError):
            class _AClass:

                @dl.DeltaMethodBlock()
                def foo(self, a) -> int:
                    return a

    def test_Interactive(self):
        with self.assertRaises(TypeError):
            @dl.Interactive({'a'})
            def foo(node):
                node.receive("a")

    def test_MigenNodeTemplate(self):
        """This test will make more sense when migen nodes are created via
        decorators.
        """
        class AMigenNode(dl.MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('a')

        with self.assertRaises(TypeError):
            AMigenNode()


class DecoratorNoOutputTypeTest(unittest.TestCase):
    """If output type is not provided in node definition Void is used."""

    def test_DeltaBlock(self):
        @dl.DeltaBlock()
        def foo(a: int):
            print(a)

        with dl.DeltaGraph():
            node = foo(1)

        self.assertEqual(node.out_type, dl.Void)

    def test_DeltaMethodBlock(self):
        class AClass:

            @dl.DeltaMethodBlock()
            def foo(self, a: int):
                print(a)

        with dl.DeltaGraph():
            node = AClass().foo(1)

        self.assertEqual(node.out_type, dl.Void)

    def test_Interactive(self):
        @dl.Interactive({"a": int})
        def foo(node):
            print(node.receive("a"))

        self.assertEqual(foo.out_type, dl.Void)

    def test_MigenNodeTemplate(self):
        """This test will make more sense when migen nodes are created via
        decorators.
        """
        class AMigenNode(dl.MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('a', dl.DOptional(int))

        self.assertEqual(AMigenNode()._out_type, dl.Void)


if __name__ == "__main__":
    unittest.main()



import logging
from queue import Empty, Queue
import sys
import threading
from time import sleep
from typing import Dict, Tuple, Type, Union

from deltalanguage.wiring import (DeltaGraph,
                                  OutPort,
                                  PyConstBody,
                                  PyFuncBody,
                                  PyInteractiveBody,
                                  PyMethodBody,
                                  PythonNode,
                                  RealNode)
from deltalanguage.logging import MessageLog, clear_loggers, make_logger

from ._queues import ConstQueue, DeltaQueue


class DeltaRuntimeExit(Exception):
    """Formal exit of
    :py:class:`DeltaGraph<deltalanguage.wiring.DeltaGraph>` will lead to a
    normal termination of _any_ runtime or runtime simulator.

    It should be raised by any of the running nodes *after* a required
    result is achieved.
    After being raised a runtime or runtime simulator will terminate
    all the processes and threads in _unspecified_ order, thus the result of
    Deltaflow program should not rely on these steps.

    Examples
    --------
    This node will stop the runtime simulator execution on receiving an input:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.DeltaBlock(allow_const=False)
        ... def foo(a: int) -> dl.Void:
        ...     print("Received result", a)
        ...     raise dl.DeltaRuntimeExit

        >>> with dl.DeltaGraph() as graph:
        ...     foo(42) # doctest:+ELLIPSIS
        foo...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        Received result 42
    """
    pass


class DeltaThread(threading.Thread):
    """Extension of ``threading.Thread``, which
    handles Deltaflow exit strategy via ``threading.excepthook``.

    Attributes
    ----------
    bad_exc : BaseException
        If a bad exception is raised by this thread it will be referred here
        for logging and afted-simulation analysis.

        If a good exception is raised it is not added here as the thread
        has finished successfully.
        There are 2 types of good exceptions:

        - ``SystemExit`` - good thread exit, i.e. the thread can be joined
          with the main thread.
        - :py:class:`DeltaRuntimeExit` - good thread exit that also signals
          to stop all the other threads via ``SystemExit`` thus the entire
          simulation will be finished.
    """

    def run(self):
        self.bad_exc = None

        try:
            if self._target:
                self._target(*self._args, **self._kwargs)
        except SystemExit:
            raise
        except DeltaRuntimeExit:
            raise
        except BaseException as e:
            self.bad_exc = e
            raise
        finally:
            del self._target, self._args, self._kwargs


class DeltaPySimulator:
    """Python runtime simulator for running
    :py:class:`DeltaGraph<deltalanguage.wiring.DeltaGraph>`.

    The main purpose of this simulator is debugging of the graph and basic
    testing. Please note that this implementation is not
    optimized for high performance, thus do not try to run big graphs in it,
    instead use Deltasimulator package.

    The advantage of this implementation is that you can debug nodes in
    Python ecosystem without need to verilate Migen code and compile any
    SystemC libraries.

    At initialization it spins out an individual ``threading.Thread``
    for each node's body. Inputs and outputs to which are managed by
    corresponding input and output ports. The transfer of data is facilitated
    by channels implemented as :py:class:`DeltaQueue`.

    Parameters
    ----------
    graph : DeltaGraph
        The graph which will be executed.
    lvl : int
        The level at which logs are displayed.
        These are the same levels as in Python's ``logging`` module.
        By default only error logs are displayed.
    msg_lvl : int
        The level at which logs from messages between nodes are displayed.
        These are the same levels as in Python's ``logging`` module.
        By default only error logs are displayed.
    switchinterval : float
        Passed to `sys.setswitchinterval`, which
        sets the interpreter’s thread switch interval (in seconds).
        This argument is provided for speed optimization of this simulator
        implementation.
    queue_size : int
        This is a global parameter for all instances of :py:class:`DeltaQueue`
        created by the simulator.
        The default value 16 is chosen in analogy with ``sc_fifo`` in SystemC,
        this means that the Python GIL will be forced to run another node's
        thread once the queue is full, thus it will prevent blockage.
    queue_interval : float
        When the simulation is called to stop, some of the queues might be
        blocking nodes at ``put`` methods as they are full. The simulator
        interrupts this at this periodicity (in seconds) and checks if
        stopping is needed.


    .. note::
        In order to reduce computation the results of constant nodes are
        evaluated only once, cached, and streamed outwards continiously.


    Examples
    --------
    After a graph is defined it is handed to this simulator:

    .. code-block:: python

        >>> import deltalanguage as dl

         # helper node that saves the result
        >>> s = dl.lib.StateSaver(int, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     s.save_and_exit(5) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 5

    Please refer to the tutorial :file:`logging_and_debugging.ipynb`
    for interactive use cases.
    """

    __quit_msg = "Quitting Delta Runtime due to error in node."

    # nodes with thread_worker
    running_body_cls: Tuple[Type[RealNode], ...] = (PyInteractiveBody,
                                                    PyFuncBody,
                                                    PyMethodBody)
    # nodes that only run once
    run_once_body_cls: Tuple[Type[RealNode], ...] = (PyConstBody,)

    def __init__(self,
                 graph: DeltaGraph,
                 lvl: int = logging.ERROR,
                 msg_lvl: int = logging.ERROR,
                 switchinterval: float = None,
                 queue_size: int = 16,
                 queue_interval: float = 1.0):
        self.log = make_logger(lvl, "DeltaPySimulator")
        self.msg_log = MessageLog(msg_lvl)
        self.set_excepthook()

        # speed optimization
        if switchinterval is not None:
            sys.setswitchinterval(switchinterval)
        self.queue_size = queue_size
        self.queue_interval = queue_interval

        # the graph
        self.graph = graph
        self.graph.do_automatic_splitting()
        self.graph.check()
        self.add_message_log()

        # i/o queues
        self.in_queues: Dict[str, Dict[str, DeltaQueue]] = {
            node.name: {}
            for node in self.graph.nodes
        }
        self.out_queues: Dict[str, Dict[str, DeltaQueue]] = {
            node.name: {}
            for node in self.graph.nodes
        }
        for node in self.graph.nodes:
            self._create_io_queues(node)

        # Signal to stop child threads
        self.sig_stop = threading.Event()

        for node in self.graph.nodes:
            node.set_communications(self)

        # child threads for node's workers
        self.threads: Dict[str, threading.Thread] = {}
        self.running = False

    def _create_io_queues(self, node):
        """Create inter-node communication queues starting from the given node.

        Depending on the type of the given node, the queues will be
        ConstQueue (for a const node) or
        DeltaQueue (the default).
        """
        for out_port in node.out_ports:
            q: DeltaQueue = self._make_queue(out_port)

            if q is not None:
                self.log.info(
                    f"creating queue: {str(out_port.port_name):_<30s} "
                    f"to {str(out_port.destination.port_name):_<30s} "
                    f"{str(out_port.port_type):_<20s}"
                )
                self.log.debug(f"It is a {type(q).__name__}. "
                               f"Optional: {q.optional}")

                # attach the queue to both nodes it connects
                self.out_queues[node.name][out_port.port_name.n_index] = q
                self.in_queues[out_port.destination.node.name][
                    out_port.destination.port_name.n_index] = q

            else:
                self.log.info(
                    f"constant queue: {str(out_port.port_name):_<30s} "
                    f"to {str(out_port.destination.port_name):_<30s} "
                    f"{str(out_port.port_type):_<20s} "
                    f"is not created"
                )

    def _make_queue(self, out_port: OutPort) -> Union[DeltaQueue, None]:
        """Decide the type of queue that should be used for an out port and
        call the constructor.

        If it is from a const node to a non-const node, it will be a
        ConstQueue. Messages between two const nodes call each other directly
        and avoid queues altogether. The default is DeltaQueue.
        """
        if isinstance(out_port.node.body, self.run_once_body_cls):
            if isinstance(out_port.destination.node.body, self.run_once_body_cls):
                return None
            return ConstQueue(out_port)

        if out_port.destination.in_port_size > 0 and self.queue_size > 0:
            maxsize = min(out_port.destination.in_port_size, self.queue_size)
        else:
            # one or both queues is 0, choose largest size
            maxsize = max(out_port.destination.in_port_size, self.queue_size)

        return DeltaQueue(out_port,
                          maxsize=maxsize,
                          queue_interval=self.queue_interval)

    def all_queues(self):
        """An iterator through all the queues.

        To avoid doublecounting iteration is done via input queues.
        """
        for queue_store in self.in_queues.values():
            for qu in queue_store.values():
                yield qu

    def add_message_log(self):
        """Set the message log for all nodes

        Parameters
        ----------
        level : int
            The new level of logging for the nodes.
        """
        for node in self.graph.nodes:
            if isinstance(node, PythonNode):
                node.set_msg_log(self.msg_log)

    def start(self):
        """Each node's worker function is started on its own thread, and
        begins listening for input. The nodes in the run-once categories get
        run now, but only if they output to a running node type.
        """
        if self.running:
            raise RuntimeError('DeltaPySimulator is already running')
        else:
            self.running = True

        try:
            for node in self.graph.nodes:
                if isinstance(node.body, self.run_once_body_cls):
                    node.run_once(self)

        except DeltaRuntimeExit:
            raise RuntimeError(
                "Constant nodes cannot raise a DeltaRuntimeExit.")
        except Exception as exc:
            raise RuntimeError(
                "Error occurred in constant node during program start."
                + "Exiting simulator.") from exc

        for node in self.graph.nodes:
            if isinstance(node.body, self.run_once_body_cls):
                continue

            elif isinstance(node.body, self.running_body_cls):
                self.log.info(f"Starting node {node.name}")
                self.threads[node.name] = DeltaThread(
                    target=node.thread_worker,
                    args=(self,),
                    name=f"Thread_{node.name}"
                )
                self.threads[node.name].start()

            else:
                self.log.error(f"node {node.name} is not of a recognised " +
                               f"class {type(node)}")

        if len(self.threads) == 0:
            raise RuntimeError("Graph cannot consist of only constant nodes.")

        # main thread is always present
        self.log.info(f"Total number of threads = {len(self.threads) + 1}")

    def run(self, timeout=None):
        """Run the simulation of the graph.

        If ``timeout`` is ``None`` runs continually until the graph calls
        :py:class:`DeltaRuntimeExit` or an error occurs.
        Otherwise it should be a floating point number specifying a timeout
        for the simulation in seconds. Note that providing the same timeout
        does not guarantee that simulation stops at the same point.
        """
        self.start()
        self.sig_stop.wait(timeout=timeout)
        self.stop()

    def set_excepthook(self):
        """Overwrite `threading.excepthook` with Deltalanguage exiting strategy.

        Strategy:

        - Raising `SystemExit` will lead to silent exit of the node's worker
          and will not signal to other workers to stop.
        - Raising :py:class:`DeltaRuntimeExit` stops the caller node's
          worker and signal to other workers to stop via `SystemExit`.
        - All other exceptions will will shut other workers via `SystemExit`,
          but the exception will be re-thrown by the simulator as a failure.
        """
        def excepthook(args):
            if args.exc_type == SystemExit:
                self.log.log(logging.DEBUG, f"Thread stopped: {args}")

            elif args.exc_type == DeltaRuntimeExit:
                self.log.log(logging.DEBUG, f"Thread stopped: {args}")
                self._stop_workers()

            else:
                self.log.log(logging.ERROR, f"Thread stopped: {args}")
                self._stop_workers()

            return

        threading.excepthook = excepthook

    def _stop_workers(self):
        if not self.sig_stop.is_set():
            self.sig_stop.set()

            # flush potential blocking queues
            for qu in self.all_queues():
                qu.flush()

    def stop(self):
        """Wait for all threads to join to join and do logging."""
        self._stop_workers()

        for th in self.threads.values():
            th.join()

        # tidy up the logs
        self.msg_log.log_messages()
        clear_loggers()
        self.running = False

        # check for bad exceptions
        for th in self.threads.values():
            if th.bad_exc:
                raise RuntimeError(
                    'At lease one exception is raised in a child thread'
                ) from th.bad_exc



import doctest
import os
import unittest


class TestDoctest(unittest.TestCase):

    def test_doctest(self):
        """Run doctest to test examples.

        Adapted from
        https://www.rosipov.com/blog/python-tests-with-doctest-and-unittest/
        """
        files = []
        root_dir = 'deltalanguage/'

        for root, _, filenames in os.walk(root_dir):
            for filename in filenames:
                if filename == '__init__.py' or filename[-3:] != '.py':
                    continue
                f = os.path.join(root, filename)
                f = f.replace('/', '.')
                f = f[:-3]
                files.append(f)

        suite = unittest.TestSuite()
        for module in files:
            suite.addTest(doctest.DocTestSuite(module))
        result = unittest.TextTestRunner(verbosity=2).run(suite)
        # If errors or failures have occured, raise the first one
        if result.errors:
            raise AssertionError(result.errors[0][1])
        if result.failures:
            raise AssertionError(result.failures[0][1])


if __name__ == "__main__":
    unittest.main()



from abc import abstractmethod
import atexit
import logging
from time import sleep
from typing import Callable, Type, Union

import migen

from deltalanguage.data_types import (BaseDeltaType,
                                      DeltaTypeError,
                                      DOptional,
                                      Void,
                                      as_delta_type,
                                      make_forked_return)
from deltalanguage.logging import make_logger
from .._body_templates import BodyTemplate
from .._node_templates import NodeTemplate
from .latency import Latency
from .node_bodies import PyMigenBody


class MigenNodeTemplate(BodyTemplate):
    """Base class for nodes that use ``migen`` for generation of internal
    logic.

    Users have to create a subclass and define :py:meth:`migen body` with
    logic identical to the native migen code.
    The only extra bit that needs to be added is the connection to input and
    output protocol adapters, which is illustrated in examples below.
    More use cases can be found in :doc:`tutorials/tutorials` and
    :doc:`examples/examples`.

    Parameters
    ----------
    tb_num_iter : int
        Number of iteration the testbench will run.
        If ``None`` will run forever.
        Note that the iteration counted, provided by migen, does
        not interact with the DeltaPySimulator counter.
    name : str
        Node name used by DeltaGraph.
    lvl : int
        Logging level.
    vcd_name : str
        Write a VCD dump of the signals inside of :py:attr:`_dut`
        by using ``migen.run_simulation``
    generics : dict
        Generic constants for the node (e.g. number of bits,
        number of pipeline stages, etc..)

    Attributes
    ----------
    _dut : migen.Module
        Migen module itself.
    _tb : Callable
        Testbench generator, this includes :py:attr:`_dut` and its wiring to
        I/O ports.
    _sim_object : migen.Simulator
        Simulated object evaluated by migen's testbench.
    module_name : str
        Module name of verilog-converted migen node.
        If ``None`` the name is obtained from the class.
        If the module is included
        in a graph, the node's index is appended to the module name.
    debug_signals : Dict[str, Tuple[migen.Signal, str]]
        Internal signal to print in the debuggin message after each iteration
        of the testbench. The key is a name, the value is the tuple with the
        signal itself and the number format, e.g. see
        https://mkaz.blog/code/python-string-format-cookbook/


    .. TODO:: Clean up user-facing methods and hide the rest.

    Examples
    --------
    This example is a mere illustration of the syntax with migen.
    Firstly, the user defines inputs and outputs of the node and
    then defines the logic using ``migen`` syntax:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> class Foo(dl.MigenNodeTemplate):
        ...
        ...     def migen_body(self, template):
        ...         # define I/O: each port gets `data`, `valid`, and `ready`
        ...         a = template.add_pa_in_port('a', DOptional(int))
        ...         b = template.add_pa_in_port('b', DOptional(int))
        ...         out = template.add_pa_out_port('out', int)
        ...
        ...         # LOGIC: identical to `migen.Module` syntax
        ...         # receive inputs
        ...         self.comb += [
        ...             a.ready.eq(1),
        ...             b.ready.eq(1),
        ...         ]
        ...         # propagate to output
        ...         self.comb += [
        ...             out.data.eq(a.data + b.data),
        ...             out.valid.eq(a.valid & b.valid),
        ...         ]

    .. note::
        All inputs of ``Foo`` are optional, this
        is a default behaviour whilst simulating hardware. Please refer to
        migen tutorials on how to define logic so as to signal processing is
        handled propely.

    Now let's connect a simple graph:

    .. code-block:: python

        >>> foo = Foo()
        >>> s = dl.lib.StateSaver(int, verbose=True)

        # note how I/O is handled
        >>> with dl.DeltaGraph() as graph:
        ...     res = foo.call(a=40, b=2)
        ...     s.save_and_exit(res.out) # doctest:+ELLIPSIS
        save_and_exit...

    And run it:

    .. code-block:: python

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 42


    """

    def __init__(self,
                 tb_num_iter: int = None,
                 name: str = None,
                 lvl: int = logging.ERROR,
                 vcd_name: str = None,
                 generics: dict = None,
                 node_template: NodeTemplate = None):
        if name is None:
            name = type(self).__name__
        super().__init__(name, Latency(clocks=1), lvl)

        self.module_name = self.name
        self.log = make_logger(lvl, f"{self.name}")
        self.vcd_name = vcd_name

        if generics is not None:
            self.generics = generics

        # Predefining values to satisfy linting check.
        self.submodules = None
        self.inputs = None
        # None for output implies there are no out-ports
        self.output = None
        self.specials = None
        self.comb = None
        self.sync = None
        self.debug_signals = {}

        self._dut = migen.Module()
        self._dut.in_ports = {}
        self._dut.out_ports = {}

        self.__class__.migen_body(self=self._dut, template=self)
        self.rename_port_signals()

        if len(self._dut.out_ports.items()) > 0:
            # Make all outputs named, as it's needed for verilog
            self._out_type, self._ForkedOutput = make_forked_return(
                {name: type_ for name, (_, type_)
                 in self._dut.out_ports.items()}
            )
        else:
            self._out_type = Void

        # define a stimulus generator
        self._tb = self.tb_generator(tb_num_iter)

        # set up a simulator and perform run-once elaboration
        self._sim_object = None
        atexit.register(self.cleanup)

        # Merge with or create a new NodeTemplate
        in_params = dict()
        for name, (_, type_) in self._dut.in_ports.items():
            in_params[name] = type_
        NodeTemplate.merge_migenblock(node_template, self, in_params,
                                      self._out_type)

    def add_pa_in_port(self, name: str, t: DOptional):
        """Add input protocol adaptor v2.

        Parameters
        ----------
        t : DOptional
            Data type must be marked with DOptional.

        Returns
        -------
        migen.Record


        .. todo::
            Possibly use DSize for data/valid/ready instead of int.
        """
        if isinstance(t, DOptional):
            df_t = t.type
        else:
            raise DeltaTypeError('Migen works only with optional inputs.\n'
                                 'Please use DOptional wrapper.')

        in_port = migen.Record([
            ("data", df_t.size.val, migen.DIR_M_TO_S),
            ("valid", 1, migen.DIR_M_TO_S),
            ("ready", 1, migen.DIR_S_TO_M)
        ])

        self._dut.in_ports[name] = (in_port, t)

        return in_port

    def add_pa_out_port(self, name: str, t: Union[Type, BaseDeltaType]):
        """Add output protocol adaptor v2.

        Parameters
        ----------
        type_ : Union[Type, BaseDeltaType]
            Data type.

        Returns
        -------
        migen.Record


        .. todo::
            Possibly use DSize for data/valid/ready instead of int.
        """
        df_t = as_delta_type(t)
        if isinstance(df_t, DOptional):
            raise DeltaTypeError('out_port cannot be DOptional')

        out_port = migen.Record([
            ("data", df_t.size.val, migen.DIR_S_TO_M),
            ("valid", 1, migen.DIR_S_TO_M),
            ("ready", 1, migen.DIR_M_TO_S)
        ])

        self._dut.out_ports[name] = (out_port, t)

        return out_port

    def cleanup(self):
        """Actions that are performed just before the object is destroyed
        At the moment we only care about closing the trace file that we have
        generated by passing vcd_name to `migen.Simulator`.

        .. note::
            There is no direct way to call `close` from a
            running `DeltaGraph`. What we know is that if we are about to
            destroy this object there are no longer active references to it,
            that indicates that the simulation has been stopped.

        .. note::
            When `vcd_name==None` the current implementation of migen returns
            a null VCDWriter that implements `close`.
        """
        if self._sim_object:
            self._sim_object.vcd.close()

    def close(self):
        self._sim.vcd.close()

    def rename_port_signals(self):
        """Rename signals used in ports"""
        for port_name, (port, _) in self._dut.in_ports.items():
            for signal_name, _, _ in port.layout:
                signal = getattr(port, signal_name)
                signal.name_override = port_name + "_in_" + signal_name

        for port_name, (port, _) in self._dut.out_ports.items():
            for signal_name, _, _ in port.layout:
                signal = getattr(port, signal_name)
                signal.name_override = port_name + "_out_" + signal_name

    @property
    def _sim(self):
        """Migen nodes can only be elaborated once, so we need to call
        Simulator ONLY if we really want to simulate the design as it will
        block creation of the serialised representation.

        This shim injects a version of _sim that is only constructed on demand.
        """
        if self._sim_object is None:
            sim_object = migen.Simulator(self._dut, self._tb,
                                         vcd_name=self.vcd_name)
            self._sim_object = sim_object
            self._elaborate()
            return sim_object
        else:
            return self._sim_object

    @abstractmethod
    def migen_body(self, template):
        """This is where user defines the combinational and synchronous logic
        that describes the node.
        """
        pass

    def construct_body(self):
        """Construct the ``PyMigenBody`` this is a template for.
        """
        return PyMigenBody(self._py_sim_body, self, self._tags)

    def tb_generator(self, tb_num_iter: int = None):
        """Generator, a.k.a. testbench.

        This method generates a sequence of stimuli for the migen module
        it may run for any number of cycles, but should ensure the input and
        output queues are eventually serviced.
        """
        tb_iter = 0
        while True:
            if tb_num_iter is not None:
                if tb_iter == tb_num_iter:
                    raise ValueError(
                        f'{self.name}.tb_generator needs more iteration to '
                        f'finish, please increase it and restart.'
                    )

            self.log.debug(f"start iteration {tb_iter}")

            # 1. Determine if all in_ports are ready, thus the module is ready
            # to receive data
            is_module_ready = True
            for _, (port, _) in self._dut.in_ports.items():
                is_port_ready = yield port.ready
                if not is_port_ready:
                    is_module_ready = False
                    break

            self.log.debug(f"all _dut.in_ports ready: {is_module_ready}")
            self.log.debug(f"inputs available: {bool(self.inputs)}")

            # 2. If yes: transfer data inputs -> in_ports
            #
            # TODO this is a quick fix before protocol adaptors are ready.
            # Currently all inputs are considered valid as long as they are
            # provided. This is not the case in general, for instance
            # the input can have all sort of junk if valid == False.
            if bool(self.inputs) and is_module_ready:
                for name, (port, _) in self._dut.in_ports.items():
                    if name in self.inputs:
                        data = self.inputs[name]
                        if data is None:
                            data = 0
                            valid = 0
                        else:
                            valid = 1
                    else:
                        raise AttributeError(
                            f'''{self.__class__.__name__} input {name}
                            is invalid''')
                    self.log.debug(f"writing in_port name={name} "
                                   f"valid={valid}, data={data}")
                    yield port.valid.eq(valid)
                    yield port.data.eq(data)

            # 3. Check conditions to run the migen body
            # TODO Currently out_ports are always ready to receive data,
            # i.e. out_ports have ready set to 1 by us.
            # This will be controlled by Runtime in the future, which will
            # set it according to the readiness of the receiving node and any
            # other extra conditions.
            for _, (port, _) in self._dut.out_ports.items():
                yield port.ready.eq(1)
            self.log.debug("out_ports set ready")

            # 4. Transfer data out_ports -> output
            #
            # If output data for a channel is not valid,
            # the corresponding output is set to None
            if len(self._dut.out_ports.items()) > 0:
                out_tmp = {}
                for name, (port, _) in self._dut.out_ports.items():
                    if (yield port.valid):
                        out_tmp[name] = yield port.data
                    else:
                        out_tmp[name] = None
                self.output = self._ForkedOutput(**out_tmp)

                self.log.debug(f"retrieved data output={self.output}")
            else:
                self.log.debug(f"no out_ports => no output")

            self.log.debug(f"end of iteration {tb_iter}")

            # 5. Evaluate one clock cycle
            yield

            # 6. Status after the cycle
            self.log.debug(f'<<<<<<<<<<< STATUS AFTER YIELD IN ITER {tb_iter}')
            self.log.debug(f'{self.inputs=}')
            self.log.debug('in_ports:')
            for name, (port, _) in self._dut.in_ports.items():
                data = yield port.data
                valid = yield port.valid
                ready = yield port.ready
                self.log.debug(f'{name=}: {data=} {valid=} {ready=}')

            self.log.debug('debug_signals:')
            for name, (signal, format_) in self.debug_signals.items():
                sig = yield signal
                self.log.debug(f'{name=}: {sig:{format_}}')

            self.log.debug('out_ports:')
            for name, (port, _) in self._dut.out_ports.items():
                data = yield port.data
                valid = yield port.valid
                ready = yield port.ready
                self.log.debug(f'{name=}: {data=} {valid=} {ready=}')
            self.log.debug(f'{self.output=}')
            self.log.debug('>>>>>>>>>>>')

            tb_iter += 1

    def _elaborate(self):
        """First part of `migen.Simulator.run`.

        This part executes the combinational logic.
        """
        self._sim.evaluator.execute(self._sim.fragment.comb)
        self._sim._commit_and_comb_propagate()

    def _clock(self):
        """One iteration of the second part of `migen.Simulator.run`.

        This part executes one clock cycle of the synchronous logic and
        propagates the combinational logic.
        """
        delta_t, rising, falling = self._sim.time.tick()
        self._sim.vcd.delay(delta_t)
        for clk_d in rising:
            self._sim.evaluator.assign(
                self._sim.fragment.clock_domains[clk_d].clk,
                1
            )
            if clk_d in self._sim.fragment.sync:
                self._sim.evaluator.execute(self._sim.fragment.sync[clk_d])
            if clk_d in self._sim.generators:
                self._sim._process_generators(clk_d)
        for clk_d in falling:
            self._sim.evaluator.assign(
                self._sim.fragment.clock_domains[clk_d].clk,
                0
            )
        self._sim._commit_and_comb_propagate()

    @staticmethod
    def _py_sim_body(self, **kwargs):
        """This method is handled over to the node body constructor, and it
        will be evaluated by the testbench at each clock cycle.
        """
        self.inputs = kwargs  # enqueue the input
        self.output = None

        # runs until the generator terminates.
        # for this impl, this will happen when inputs is empty
        self._clock()

        return self.output

    def call(self, *args, **kwargs):
        """This performs the same action as ``Interactive.call``,
        i.e. use this method to provide inputs to the node while building
        the graph.

        This implementation uses the global clock of `DeltaPySimulator`,
        at each tick:

            - it takes the input from the in_queues
            - evaluates the node
            - sends output to out_queues

        What we want eventually is that migen module is independent of
        `DeltaPySimulator`, more specifically:

            - migen module connects to input / output `DeltaQueue`
            - it clocks with its own clock form migen, i.e. it is started by
              `DeltaPySimulator` only once

        Parameters
        ----------
        args
            Non-keyworded arguments passed to the testbench.
        kwargs
            Keyworded arguments passed to the testbench.


        .. todo::
            `args` are being ignored => raise an error if passed.
        """

        from .._delta_graph import DeltaGraph
        if DeltaGraph.stack():
            graph = DeltaGraph.current_graph()
            retval = self._call(graph, *args, **kwargs)
            self.module_name = retval.name
        else:
            # it does only one clock, so you need to place it in a loop
            retval = self._py_sim_body(self, **kwargs)
        return retval

    @staticmethod
    def unpack_record(port):
        """Helper function to access `migen.Signal` corresponding to the port.
        """
        return (getattr(port, name) for name, _, _ in port.layout)

    def get_serialised_body(self) -> Union[str, Callable]:
        """This function returns a object to be included in a .df program.

        Currently, it will return a fully-specified verilog string, but in
        future it may return a partial function with kwargs for template
        parameters.
        """
        fully_parameterised = True
        if fully_parameterised:
            ios = set()

            for _, (port, _) in self._dut.in_ports.items():
                for signal in self.unpack_record(port):
                    ios.add(signal)

            for _, (port, _) in self._dut.out_ports.items():
                for signal in self.unpack_record(port):
                    ios.add(signal)

            v_str = migen.fhdl.verilog.convert(
                self._dut, ios, name=self.module_name)
            return v_str

        else:
            raise NotImplementedError



import unittest
from collections import OrderedDict

from deltalanguage._utils import NamespacedName
from deltalanguage.data_types import (DBool, DInt, DOptional, DeltaTypeError,
                                      Void, make_forked_return)
from deltalanguage.runtime import DeltaPySimulator, DeltaRuntimeExit
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaMethodBlock,
                                  DeltaGraph,
                                  ForkedNode,
                                  Interactive,
                                  placeholder_node_factory,
                                  NodeTemplate,
                                  MigenNodeTemplate,
                                  PyMigenBody,
                                  RealNode)


class NoBodyNodeTest(unittest.TestCase):
    """Tests regarding using `NodeTemplate` to create a node with no bodies
    """

    def setUp(self):
        def func_for_placeholder() -> int:
            return 7
        self.func_for_placeholder = func_for_placeholder

    def test_forked_input(self):
        ForkedReturnT, ForkedReturn = make_forked_return({'a': int, 'b': bool})
        template_2 = NodeTemplate(in_params={'a': int, 'b': bool})

        @DeltaBlock()
        def add_1_true(n: int) -> ForkedReturnT:
            return ForkedReturn(a=n+1, b=True)

        with DeltaGraph() as graph:
            forked_input = add_1_true(n=5)
            template_2.call(a=forked_input.a, b=forked_input.b)

        self.assertTrue(graph.check())

    def test_placeholder_input(self):
        template_3 = NodeTemplate(in_params=OrderedDict([('a', int)]))

        with DeltaGraph() as graph:
            a = placeholder_node_factory()
            template_3.call(a=a)

        a.specify_by_func(self.func_for_placeholder)
        self.assertTrue(graph.check())

    def test_partial_arg_types(self):
        template_4 = NodeTemplate(in_params={'a': int, 'b': bool})

        @DeltaBlock()
        def bool_and(a: bool, b: bool) -> bool:
            return a and b
        with DeltaGraph() as graph:
            a = placeholder_node_factory()
            template_4.call(a=a,
                            b=bool_and(a=True, b=False))

        a.specify_by_func(self.func_for_placeholder)
        self.assertTrue(graph.check())

    def test_template_input_raises_exception(self):
        @DeltaBlock()
        def add(a: int, b: int) -> int:
            return a+b

        template_1 = NodeTemplate(in_params={'a': bool})

        with self.assertRaises(DeltaTypeError):
            with DeltaGraph() as graph:
                template_1.call(a=add(a=1, b=2))
            graph.check()


class OpCacher2():

    def __init__(self):
        self._add_cache = {}

    @DeltaMethodBlock()
    def cached_add(self, a: int, b: int) -> int:
        if (a, b) in self._add_cache:
            return self._add_cache[(a, b)]
        else:
            result = a+b
            self._add_cache[(a, b)] = result
            return result


class NoBodyAddBody(unittest.TestCase):
    """Tests for when the default call is used to create a node with no bodies
    and then ``add_body`` is used to give the node a body. 
    """

    def test_add_func(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(allow_const=False)
        def simple_add(a: int, b: int) -> int:
            return a + b

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 0)
        n1.add_body(simple_add)
        self.assertEqual(len(n1.bodies), 1)
        self.assertIn('simple_add', n1.body.access_tags)

    def test_add_method(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)
        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 0)
        n1.add_body(OpCacher2().cached_add)
        self.assertEqual(len(n1.bodies), 1)
        self.assertIn('cached_add', n1.body.access_tags)

    def test_add_multiple(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(allow_const=False)
        def simple_add(a: int, b: int) -> int:
            return a + b

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 0)
        n1.add_body(OpCacher2().cached_add)
        n1.add_body(simple_add)
        self.assertEqual(len(n1.bodies), 2)
        self.assertIn('cached_add', n1.body.access_tags)

    def test_add_interactive(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @Interactive(in_params=OrderedDict([('a', int), ('b', int)]),
                     out_type=int)
        def broken_adder(node: RealNode):
            a = node.receive('a')
            b = node.receive('b')
            node.send(a+b+1)

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        self.assertEqual(len(n1.bodies), 0)
        n1.add_body(broken_adder)
        self.assertEqual(len(n1.bodies), 1)
        self.assertIn('broken_adder', n1.body.access_tags)

    def test_add_migen(self):
        test_template1 = NodeTemplate(
            name="test_1",
            in_params={'a': DOptional(int), 'b': DOptional(int)}
        )

        class AMigenNode(MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('a', DOptional(int))
                template.add_pa_in_port('b', DOptional(int))

        with DeltaGraph():
            n1 = test_template1.call(a=2, b=3)

        self.assertEqual(len(n1.bodies), 0)
        n1.add_body(AMigenNode())
        self.assertEqual(len(n1.bodies), 1)
        self.assertIn(PyMigenBody, n1.body.access_tags)

    def test_add_invalid_body(self):
        test_template1 = NodeTemplate(name="test_1",
                                      in_params=OrderedDict(
                                          [('a', int), ('b', int)]),
                                      out_type=int)

        @DeltaBlock(allow_const=False)
        def simple_add_to_bool(a: int, b: int) -> bool:
            return bool(a + b)

        with DeltaGraph():
            n1 = test_template1.call(2, 3)

        with self.assertRaises(ValueError):
            n1.add_body(simple_add_to_bool)


if __name__ == "__main__":
    unittest.main()



"""Summary information about the package
    Attributes
    ----------
    __license__ : str
        Short description of our license.
    __copyright__ : str
        Copyright notice
    __url__ : str
        Project URL
    __contributors__ : str
        Single-line list of all project contributors.
    __contributors_lines__ : str
        Multi-line list of all project contributors.
    __email__ : str
        Contact email for the project.
    __version__ : str
        Package version in format ``<major>.<minor>.<micro>``
    __version_description__ : str
        Short description for the current version.
    __short_description__ : str
        Short summary of the package.
    __doc__ : Str
        Longer description of the package.
    __platforms__ : List[str]
        List of supported platforms
"""

__all__ = [
    "__license__",
    "__copyright__",
    "__url__",
    "__download_url__",
    "__docs_url__",
    "__contributors__",
    "__contributors_lines__",
    "__email__",
    "__version__",
    "__version_description__",
    "__short_description__",
    "__doc__",
    "__platforms__"
]

__license__ = "MIT License with Commons Clause"
__copyright__ = "2020, River Lane Research Ltd"


# Source URL
__url__ = "https://github.com/riverlane/deltalanguage"
# Package Hosting URL
__download_url__ = "https://pypi.org/project/deltalanguage/#files"
# Docs Hosting Url
__docs_url__ = "https://riverlane.github.io/deltalanguage"

contributors = ["Kenton Barnes",
                "James Bate",
                "Anton Buyskikh",
                "Laurin Fischer",
                "Marco Ghibaudi",
                "Gianmarco Girau",
                "Alex Moylett",
                "Tom Parks",
                "Brendan Reid",
                "Dan Seremet",
                "Jan Snoeijs",
                "Robin Sterling",
                "Dan Underwood",
                "Noon Van Der Silk",
                "Maria Violaris"]
__contributors__ = ", ".join(contributors)
__contributors_lines__ = "\n".join(contributors)
__email__ = "deltaflow@riverlane.com"

version_info = (0, 5, 1)
"""Tuple[int, int, int] : version information
The three components of the version:
``major``, ``minor`` and ``micro``: Module level variable documented inline.
"""
__version__ = ".".join(map(str, version_info[:3]))
__version_description__ = "Deltalangauage MVP Public Release."


__short_description__ = "Hosted programming language for Deltaflow"
__doc__ = """
    Deltalangauage allows the programmer to define a graph of all the compute
    elements and declare the transport of data between them. Depending on which
    type of hardware a part of the program is running on, the nodes of this
    graph get filled with python (for the CPU) or migen (for the FPGAs).
    A hardware abstraction layer is built into the language, that standardises
    which operations can be triggered on qubits.
    This makes Deltaflow portable across technologies.
    """

__platforms__ = ['Ubuntu 20.04']



"""Deltaflow specific exceptions."""


class DeltaTypeError(Exception):
    """Raised on multiple occasions:

      - if Deltaflow data type is mismanaged in general
      - if a connection of
        :py:class:`DeltaGraph<deltalanguage.wiring.DeltaGraph>`
        has a mismatch at the receiving and sending ends
    """
    pass


class DeltaIOError(Exception):
    """Raised when :py:class:`DeltaGraph<deltalanguage.wiring.DeltaGraph>` has
    a an error in connectivity.
    """
    pass



"""
Primitives for examples, tests, and just general use.
"""
import json
from copy import deepcopy
from typing import Callable, Iterable, Type, Union

from ..data_types import (BaseDeltaType, Void,
                          delta_type, make_forked_return,
                          DComplex, DArray, DRecord, DTuple)
from ..runtime import DeltaRuntimeExit
from ..wiring import (DeltaBlock,
                      DeltaMethodBlock,
                      Interactive,
                      PythonNode,
                      InteractiveBodyTemplate)


def make_generator(val: Union[object, Iterable],
                   reps: int = None,
                   verbose: bool = False) -> InteractiveBodyTemplate:
    """Used to create a generator node that
    produces a series of messages of the same data type.

    This generator is useful for checking the robustness of the graph vs
    multiple messages and soak testing in general. Also used for making a
    testbench for a node under design.

    Parameters
    ----------
    val : Union[object, Iterable]
        The output value(s) of the generator.
        Its Deltaflow type is recognized automatically by
        :py:class:`delta_type<deltalanguage.data_types.delta_type>`.
        See ``reps`` for explanation.
    reps : int
        Number of repeated messages.
        If ``reps`` is ``None`` then ``val`` must be iterable.
        If ``reps`` is an integer then `val` is sent out this number of times.
    verbose : bool
        If ``True`` prints the status.

    Examples
    --------
    In this example you can find a basic use of this function.
    Note that the generator should be used in the same way as an
    interactive node:

    .. code-block:: python

        >>> import deltalanguage as dl

        # Generate 5 integers
        >>> generator = dl.lib.make_generator(10, reps=5)

        # Receive 5 integers and send their sum
        >>> @dl.Interactive({"a": int}, int)
        ... def accumulator(node):
        ...     memory = []
        ...     for _ in range(5):
        ...         input = node.receive("a")
        ...         memory.append(input)
        ...     node.send(sum(memory))

        # Save the result
        >>> s = dl.lib.StateSaver(int, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     generator_out = generator.call()
        ...     accumulator_out = accumulator.call(a=generator_out)
        ...     s.save_and_exit(accumulator_out) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 50

    Also see this alternative usage:

    .. code-block:: python

        >>> generator = dl.lib.make_generator([1, 2, 3, 4, 5])

        >>> with dl.DeltaGraph() as graph:
        ...     generator_out = generator.call()
        ...     accumulator_out = accumulator.call(a=generator_out)
        ...     s.save_and_exit(accumulator_out) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 15
    """
    if reps is None:
        if not isinstance(val, Iterable):
            raise ValueError('If reps is None, then val must be iterable')

        elem_type = delta_type(val[0])
        if not all(elem_type == delta_type(e) for e in val):
            raise TypeError('Elements of val should be of the same type')

        vals_to_send = val
    else:
        elem_type = delta_type(val)
        vals_to_send = (deepcopy(val) for _ in range(reps))

    @Interactive({}, elem_type)
    def generator(node: PythonNode):
        for v in vals_to_send:
            if verbose:
                print(f"sending {val}")

            node.send(v)

    return generator


def make_splitter(t: Union[Type, BaseDeltaType],
                  reps: int,
                  allow_const=True) -> Callable:
    """Used to create a splitter node, which sends multiple copies of
    incoming messages to different nodes via individual outputs.

    Names of output ports go in ascending order as ``out0``, ``out1``,
    ``out2``, etc. The total number of outputs is equal to ``reps``.

    Examples
    --------
    In this example you can find a basic use of this function:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.DeltaBlock()
        ... def adder(a: int, b: int) -> int:
        ...     return a + b

        >>> state_saver = dl.lib.StateSaver(int, verbose=True)

        >>> splitter = dl.lib.make_splitter(int, 2)

        >>> with dl.DeltaGraph() as graph:
        ...     splitter_out = splitter(10)
        ...     adder_out = adder(splitter_out.out0, splitter_out.out1)
        ...     state_saver.save_and_exit(adder_out) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 20
    """
    _SplitterT, _SplitterC = make_forked_return(
        {'out' + str(i): t for i in range(reps)}
    )

    @DeltaBlock(name="splitter", allow_const=allow_const)
    def _splitter(to_split: t) -> _SplitterT:
        return _SplitterC(*(deepcopy(to_split) for _ in range(reps)))

    return _splitter


class DeltaJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        obj_type = delta_type(obj)
        if isinstance(obj_type, DComplex):
            return {"real": obj.real, "imaginary": obj.imag}
        elif isinstance(obj_type, (DArray, DTuple)):
            return [self.default(o) for o in obj]
        elif isinstance(obj_type, DRecord):
            props = obj.__annotations__.keys()
            d = {p: self.default(getattr(obj, p)) for p in props}
            return d

        # Assume it's encodable; if we go back to JSONEncoder.default
        # this just returns a type error. See:
        # https://docs.python.org/3/library/json.html#json.JSONEncoder.default
        return obj


class StateSaver:
    """Class used for saving states via various methods.

    This is useful for debugging and testing purposes, in particular
    users can store and benchmark data transmitted between nodes.
    Also used for making a testbench for a node under design.

    Parameters
    ----------
    t : Union[Type, BaseDeltaType]
        Type of messages, by default object.
    condition : Callable
        Used for the conditional blocks, see examples.
    verbose : bool
        If ``True`` prints a status on node's activation.
    filename : str
        A filepath in which to save results, incrementally.

        If provided, this file will be appended with
        line-based JSON representation of the data saved.
        Note: The writing to the file is done when
        ``save`` is called is called.

    Examples
    --------
    In this graph a single message is saved via ``save``:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> s = dl.lib.StateSaver(int, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     s.save(42) # doctest:+ELLIPSIS
        save...

    Same as ``save`` but followed by the formal graph exit:

    .. code-block:: python

        >>> with dl.DeltaGraph() as graph:
        ...     s.save_and_exit(42) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 42

    Same as ``save_and_exit`` but if a condition is
    fulfilled (in this case it is):

    .. code-block:: python

        >>> s = dl.lib.StateSaver(int, condition=lambda x: x>10, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     s.save_and_exit_if(42) # doctest:+ELLIPSIS
        save_and_exit_if...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 42

    Same as ``save`` but also returns a status, which
    can be used to trigger futher actions in the graph:

    .. code-block:: python

        >>> gen = dl.lib.make_generator([42])
        >>> s0 = dl.lib.StateSaver(int, verbose=True)
        >>> s1 = dl.lib.StateSaver(bool, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     status = s0.save_and_ack(gen.call())
        ...     s1.save_and_exit(status) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 42
        saving True

    Also we can just save a message on passing:

    .. code-block:: python

        >>> gen = dl.lib.make_generator([42])
        >>> s0 = dl.lib.StateSaver(int, verbose=True)
        >>> s1 = dl.lib.StateSaver(int, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     message = s0.transfer(gen.call())
        ...     s1.save_and_exit(message) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 42
        saving 42

    And add a condition:

    .. code-block:: python

        >>> gen = dl.lib.make_generator([3, 8])
        >>> s0 = dl.lib.StateSaver(int, condition=lambda x: x%2==0,
        ...                        verbose=True)
        >>> s1 = dl.lib.StateSaver(int, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     message = s0.transfer_if(gen.call())
        ...     s1.save_and_exit(message) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 3
        saving 8
        saving 8

    .. warning::
        While ``t`` is optional, the default value of ``object`` should only
        be used for debugging in the Python simulator. Not specifying a type
        will cause execution in Deltasimulator to fail.
    """

    def __init__(self, t: Union[Type, BaseDeltaType] = object,
                 condition=None,
                 verbose=False,
                 filename=None):
        self.saved = []
        self.condition = condition
        self.verbose = verbose
        self.filename = filename

        @DeltaBlock(allow_const=False)
        def save(val: t) -> Void:
            """Save input of any type.

            Parameters
            ----------
            val : object
                Incoming message.
            """
            self.store(val)
        self.save = save

        @DeltaBlock(allow_const=False)
        def save_and_exit(val: t) -> Void:
            """Save input of any type and exit.

            Parameters
            ----------
            val : object
                Incoming message.
            """
            self.store(val)
            raise DeltaRuntimeExit
        self.save_and_exit = save_and_exit

        @DeltaBlock(allow_const=False)
        def save_and_exit_if(val: t) -> Void:
            """Save input of any type and exit if the condition is fulfilled.

            Parameters
            ----------
            val : object
                Incoming message.
            """
            self.store(val)
            assert self.condition is not None, "Undefined condition"
            if self.condition(val):
                raise DeltaRuntimeExit
        self.save_and_exit_if = save_and_exit_if

        @DeltaBlock(allow_const=False)
        def save_and_ack(val: t) -> bool:
            """Save input of any type and return an acknowledgement.

            Parameters
            ----------
            val : object
                Incoming message.

            Returns
            -------
            bool
            """
            self.store(val)
            return True
        self.save_and_ack = save_and_ack

        @DeltaBlock(allow_const=False)
        def transfer(val: t) -> t:
            """Save input of any type and return it.

            Parameters
            ----------
            val : object
                Incoming message.

            Returns
            -------
            object
            """
            self.store(val)
            return val
        self.transfer = transfer

        @DeltaBlock(allow_const=False)
        def transfer_if(val: t) -> t:
            """Save input of any type and return it if the contition
            is fulfilled.

            Parameters
            ----------
            val : object
                Incoming message.

            Returns
            -------
            object
            """
            self.store(val)
            assert self.condition is not None, "Undefined condition"
            if self.condition(val):
                return val
        self.transfer_if = transfer_if

    def reset(self):
        """Remove all stored states."""
        self.saved = []

    def store(self, val):
        """Helper method used for storing."""
        if self.verbose:
            print(f"saving {val}")
        self.saved.append(val)

        if self.filename is not None:
            with open(self.filename, "a") as f:
                f.write(json.dumps(val, cls=DeltaJsonEncoder) + "\n")



import unittest
from unittest.mock import Mock

from test._utils import TwoIntsT, TwoInts

from deltalanguage.data_types import make_forked_return
from deltalanguage.wiring import (PyConstBody,
                                  PyFuncBody,
                                  PyMethodBody)


class MockCallback(Mock):
    
    @property
    def __name__(self):
        return "mock_callback"

class TestPyConstBody(unittest.TestCase):
    """Unit tests for PyConstBody."""

    def setUp(self):
        self.mock_callback = MockCallback()

    def test_callback(self):
        """Callback should be called."""
        node_body = PyConstBody(self.mock_callback)
        node_body.eval()
        self.mock_callback.assert_called()

    def test_return_value(self):
        """Evaluated value should be returned."""
        self.mock_callback.return_value = 3
        node_body = PyConstBody(self.mock_callback)
        self.assertEqual(3, node_body.eval())

    def test_evaluated_once(self):
        """Callback should be called only once, even if evaluated multiple
        times.
        """
        node_body = PyConstBody(self.mock_callback)
        node_body.eval()
        node_body.eval()
        self.mock_callback.assert_called_once()

    def test_args_eval(self):
        """Arguments should be evaluated and used."""
        mock_arg = MockCallback()
        mock_arg.body.eval.return_value = 5
        mock_arg.index = None
        node_body = PyConstBody(self.mock_callback, mock_arg)
        node_body.eval()
        mock_arg.body.eval.assert_called()
        self.mock_callback.assert_called_with(5)

    def test_args_evaluated_once(self):
        """Arguments should be evaluated and used only once."""
        mock_arg = MockCallback()
        mock_arg.body.eval.return_value = 5
        mock_arg.index = None
        node_body = PyConstBody(self.mock_callback, mock_arg)
        node_body.eval()
        node_body.eval()
        mock_arg.body.eval.assert_called_once()

    def test_forked_args(self):
        """If argument node gives forked return, only used indexed argument."""
        mock_arg = MockCallback()
        mock_arg.body.eval.return_value = TwoInts(x=5, y=6)
        mock_arg.index = 'x'
        node_body = PyConstBody(self.mock_callback, mock_arg)
        node_body.eval()
        self.mock_callback.assert_called_with(5)

    def test_kwargs_eval(self):
        """Keyword arguments should be evaluated and used."""
        mock_kwarg = MockCallback()
        mock_kwarg.body.eval.return_value = 5
        mock_kwarg.index = None
        node_body = PyConstBody(self.mock_callback, a=mock_kwarg)
        node_body.eval()
        mock_kwarg.body.eval.assert_called()
        self.mock_callback.assert_called_with(a=5)

    def test_kwargs_evaluated_once(self):
        """Keyword arguments should be evaluated and used only once."""
        mock_kwarg = MockCallback()
        mock_kwarg.body.eval.return_value = 5
        mock_kwarg.index = None
        node_body = PyConstBody(self.mock_callback, a=mock_kwarg)
        node_body.eval()
        node_body.eval()
        mock_kwarg.body.eval.assert_called_once()

    def test_forked_kwargs(self):
        """If keyword argument node gives forked return, only used indexed
        argument.
        """
        mock_kwarg = MockCallback()
        mock_kwarg.body.eval.return_value = TwoInts(x=5, y=6)
        mock_kwarg.index = 'x'
        node_body = PyConstBody(self.mock_callback, a=mock_kwarg)
        node_body.eval()
        self.mock_callback.assert_called_with(a=5)

    def test_globals(self):
        """Setting up globals in the environment."""
        a = 3

        def add_2():
            return a+2
        node_body = PyConstBody(add_2)
        self.assertEqual(5, node_body.eval())

    def test_receive_none_arg_nonoptional(self):
        """None should not be received for args."""
        mock_arg = MockCallback()
        mock_arg.body.eval.return_value = None
        mock_arg.index = None
        node_body = PyConstBody(self.mock_callback, mock_arg)
        with self.assertRaises(ValueError):
            node_body.eval()

    def test_receive_none_forked_arg_nonoptional(self):
        """None should not be received for args."""
        mock_arg = MockCallback()
        mock_arg.body.eval.return_value = TwoInts(None, 3)
        mock_arg.index = 'x'
        mock_arg.out_type = TwoIntsT
        node_body = PyConstBody(self.mock_callback, mock_arg)
        with self.assertRaises(ValueError):
            node_body.eval()

    def test_receive_none_kwarg_nonoptional(self):
        """None should not be received for kwargs."""
        mock_kwarg = MockCallback()
        mock_kwarg.body.eval.return_value = None
        mock_kwarg.index = None
        node_body = PyConstBody(self.mock_callback, a=mock_kwarg)
        with self.assertRaises(ValueError):
            node_body.eval()

    def test_receive_none_forked_kwarg_nonoptional(self):
        """None should not be received for kwargs."""
        mock_kwarg = MockCallback()
        mock_kwarg.body.eval.return_value = TwoInts(None, 3)
        mock_kwarg.index = 'x'
        mock_kwarg.out_type = TwoIntsT
        node_body = PyConstBody(self.mock_callback, a=mock_kwarg)
        with self.assertRaises(ValueError):
            node_body.eval()


class TestPyFuncBody(unittest.TestCase):

    def test_callback(self):
        """Test function callback is called correctly."""
        mock_callback = MockCallback()
        node_body = PyFuncBody(mock_callback)
        node_body.eval(1, 2, 3, a=4, b=5)
        mock_callback.assert_called_with(1, 2, 3, a=4, b=5)

    def test_return(self):
        """Return value should be correct."""
        mock_callback = MockCallback()
        mock_callback.return_value = 6
        node_body = PyFuncBody(mock_callback)
        self.assertEqual(node_body.eval(1, 2, 3, a=4, b=5), 6)

    def test_globals(self):
        """Setting up global variables in environment."""
        a = 5

        def return_global():
            return a+1

        node_body = PyFuncBody(return_global)
        self.assertEqual(node_body.eval(), 6)


class TestPyMethodBody(unittest.TestCase):

    def test_method_callback(self):
        """Test method callback is called correctly."""
        mock_object = MockCallback()
        node_body = PyMethodBody(mock_object.mock_callback, mock_object)
        node_body.eval(1, 2, 3, a=4, b=5)
        mock_object.mock_callback.assert_called_with(
            mock_object, 1, 2, 3, a=4, b=5)

    def test_method_globals(self):
        """Adding globals."""

        a = 12

        class TestClass:
            def add_5(self):
                return a+5
        test_object = TestClass()
        node_body = PyMethodBody(
            TestClass.add_5, test_object)
        self.assertEqual(17, node_body.eval())


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
import logging
from typing import TYPE_CHECKING, Callable, List, Dict, Type, Optional, Union

from deltalanguage.data_types import BaseDeltaType, DOptional

from ._node_classes.latency import Latency
from ._node_classes.node_bodies import (PyFuncBody,
                                        PyInteractiveBody,
                                        PyMethodBody,
                                        PyConstBody)
if TYPE_CHECKING:
    from ._node_classes.real_nodes import PythonNode


class BodyTemplate(ABC):
    """Body templates are responsible for storing some of the data for node 
    construction, specifically the data for an individual constructor.
    The most important piece of info of this type is the body
    and how to create it. 

    While :py:class:`NodeTemplate` contains enough information to create a 
    body-less node, ``NodeTemplate`` with one or more ``BodyTemplates`` is able
    to create a node with bodies.

    Parameters
    ----------
    name : str
        Name specific to this constructor
    latency : Latency
        Latency for the body this BodyTemplate will construct
    tags : List[str]
        List of user defined tags for the body this BodyTemplate will construct
    lvl : int
        The level at which logs are displayed.
    """

    def __init__(self,
                 name: str,
                 latency: Latency,
                 lvl: int = logging.ERROR,
                 tags: List[str] = []):
        self.name = name
        self.latency = latency
        self.lvl = lvl
        self._tags = tags

        self.construction_ready = True
        self.allow_const = False

        self._node_template = None
        self._graph = None

    def set_node_template(self, node_template: 'NodeTemplate'):
        """A ``BodyTemplate`` needs to have a parent NodeTemplate so it can
        possibly be linked up to other ``BodyTemplates`` that construct similar
        bodies.

        Parameters
        ----------
        node_template : NodeTemplate
        """
        self._node_template = node_template

    @property
    def out_type(self) -> Type:
        return self._node_template.out_type

    @property
    def in_params(self) -> Dict[str, Type]:
        return self._node_template.in_params

    @property
    def in_port_size(self) -> int:
        return self._node_template.in_port_size

    @property
    def node_key(self) -> Optional[str]:
        return self._node_template.node_key

    def compatible(self,
                   node_key: Optional[str],
                   in_port_size: int,
                   in_params: Dict[str, Union[BaseDeltaType, DOptional]],
                   out_type: Union[BaseDeltaType, DOptional]) -> bool:
        """Checks compatibility between the ``NodeTemplate`` of this 
        ``BodyTemplate`` and some params that are important for node creation.

        Returns
        -------
        bool
            ``True`` if paramaters are compatible with this ``NodeTemplate``
        """
        return self._node_template.compatible(node_key, in_port_size,
                                              in_params, out_type)

    @abstractmethod
    def construct_body(self):
        """Construct the body that this ``BodyTemplate`` is a template for.
        """
        pass

    def construct_const_body(self):
        """If possible, create the constant version of the body this
        ``BodyTemplate`` is a template for, else raise ``AttributeError``.
        """
        raise AttributeError("Const body construction attempted on body "
                             "template that does not support const body")

    def _call(self, graph, *args, **kwargs):
        body = self.construct_body()
        return self._node_template._standardised_call(graph, self.name,
                                                      self.lvl, self, body,
                                                      *args, **kwargs)

    def call(self, *args, **kwargs):
        """Construct a node using this ``BodyTemplate`` and its associated 
        :py:class:`NodeTemplate`.

        Parameters
        ----------
        args
            Input nodes as a list.
        kwargs
            Input nodes specified by keyword.
        """
        from ._delta_graph import DeltaGraph
        if DeltaGraph.stack():
            graph = DeltaGraph.current_graph()
            return self._call(graph, *args, **kwargs)
        else:
            raise ValueError("Node templte called when no graph was active.")

    def call_with_graph(self, graph, *args, **kwargs):
        """Construct a node using this ``BodyTemplate`` and its associated 
        :py:class:`NodeTemplate` using a non-standard (not in the stack) graph.

        Parameters
        ----------
        graph : DeltaGraph
            The graph the new node will be attached to
        args
            Input nodes as a list.
        kwargs
            Input nodes specified by keyword.
        """
        return self._call(graph, *args, **kwargs)


class FuncBodyTemplate(BodyTemplate):
    """Body template for nodes with a :py:class:`PyFuncBody`

    Parameters
    ----------
    name : str
        Name specific to this constructor.
    latency : Latency
        Latency for the body this BodyTemplate will construct.
    fn : Callable
        Function for the bodies this template creates.
    allow_const : bool
        Is the body this template produces allowed to be created as a 
        constant body.
    tags : List[str]
        List of user defined tags for the body this BodyTemplate will
        construct.
    """

    def __init__(self,
                 name: str,
                 latency: Latency,
                 lvl: int,
                 fn: Callable,
                 allow_const: bool,
                 tags: List[str] = []):
        super().__init__(name, latency, lvl, tags)
        self._callback = fn
        self.allow_const = allow_const

    def construct_const_body(self, *pos_in_nodes, **kw_in_nodes):
        """If allowed, create the constant version :py:class:`PyConstBody` 
        version of this function body
        """
        if not self.allow_const:
            raise AttributeError("Const body construction attempted on sub-"
                                 "template that does not support const body")
        else:
            return PyConstBody(self._callback, *pos_in_nodes, **kw_in_nodes)

    def construct_body(self):
        """Construct the :py:class:`PyFuncBody` that this ``BodyTemplate`` is
        a template for
        """
        return PyFuncBody(self._callback, self.latency, self._tags)


class MethodBodyTemplate(FuncBodyTemplate):
    """Body template for nodes with a :py:class:`PyMethodBody`

    Parameters
    ----------
    name : str
        Name specific to this constructor.
    latency : Latency
        Latency for the body this BodyTemplate will construct.
    fn : Callable
        Function for the bodies this template creates.
    tags : List[str]
        List of user defined tags for the body this BodyTemplate will
        construct.
    """

    def __init__(self,
                 name: str,
                 latency: Latency,
                 lvl: int,
                 fn: Callable,
                 tags: List[str] = []):
        super().__init__(name, latency, lvl, fn, False, tags)
        self.construction_ready = False

    def construct_body(self, obj):
        """Construct the :py:class:`PyMethodBody` that this ``BodyTemplate``
        is a template for.
        """
        return PyMethodBody(self._callback, obj, self.latency, self._tags)

    def _call(self, graph, obj, *args, **kwargs):
        body = self.construct_body(obj)
        return self._node_template._standardised_call(graph, self.name,
                                                      self.lvl, self, body,
                                                      *args, **kwargs)


class InteractiveBodyTemplate(BodyTemplate):
    """Body template for nodes with a :py:class:`PyInteractiveBody`

    Parameters
    ----------
    name : str
        Name specific to this constructor
    latency : Latency
        Latency for the body this BodyTemplate will construct.
    proc : Callable[[PythonNode], None]
        Process for the bodies this template creates.
    tags : List[str]
        List of user defined tags for the body this BodyTemplate will
        construct.
    """

    def __init__(self,
                 name: str,
                 latency: Latency,
                 lvl: int,
                 proc: Callable[['PythonNode'], None],
                 tags: List[str] = []):
        super().__init__(name, latency, lvl, tags)
        self._proc = proc

    def construct_body(self):
        """Construct the :py:class:`PyInteractiveBody` that this 
        ``BodyTemplate`` is a template for.
        """
        return PyInteractiveBody(self._proc, self._tags)

    def _call(self, graph, *args, **kwargs):
        if args:
            raise ValueError("Please use keywords for interactive node "
                             "input specification.")
        body = self.construct_body()
        return self._node_template._standardised_call(graph, self.name,
                                                      self.lvl, self, body,
                                                      *args, **kwargs)



from abc import ABC, abstractclassmethod

from deltalanguage.data_types import DUInt, DSize


class IQuantumSimulator(ABC):
    """Abstract class for interfacing with a quantum simulators that interacts
    with the HAL.
    """

    @abstractclassmethod
    def accept_command(
        cls,
        command: DUInt(DSize(32))
    ) -> DUInt(DSize(32)):
        """Performs required logic based on received commands, and returns
        results if command is a measurement.

        Parameters
        ----------
        command : DUInt(DSize(32))
            The HAL command to deconstruct and use to perform actions.

        Returns
        -------
        DUInt(DSize(32))
            Result of a measurement command.
        """
        pass



from ._commands import (command_creator,
                        Masks,
                        measurement_unpacker,
                        Opcode,
                        Shifts)
from ._hardware_abstraction_layer_node import (HardwareAbstractionLayerNode,
                                               hal_template)



"""User facing decorators for building DeltaGraph nodes."""

from __future__ import annotations
from functools import wraps
import logging
from typing import (TYPE_CHECKING,
                    Callable,
                    Dict,
                    Optional,
                    OrderedDict,
                    Tuple,
                    Type,
                    Union)

import attr

from deltalanguage.data_types import (BaseDeltaType,
                                      DeltaTypeError,
                                      DOptional,
                                      ForkedReturn,
                                      as_delta_type,
                                      Void)
from ._delta_graph import DeltaGraph
from ._node_classes.real_nodes import (as_node,
                                       get_func_in_params_out_type,
                                       PythonNode)
from ._node_classes.node_bodies import PyInteractiveBody
from ._node_templates import NodeTemplate, InteractiveBodyTemplate

if TYPE_CHECKING:
    from deltalanguage.wiring import Latency


def DeltaBlock(
    template: NodeTemplate = None,
    allow_const: bool = True,
    node_key: Optional[str] = None,
    name: str = None,
    in_port_size: int = 0,
    latency: Latency = None,
    lvl: int = logging.ERROR
):
    """Decorator to turn a function to a block for use in
    :py:class:`DeltaGraph`.

    If called in the context of :py:class:`DeltaGraph` it will
    return a stateless node, which means that the output of such a node is
    fully determined by its inputs; exactly as it happens in functional
    programming.

    The node is evaluated when all compulsory inputs are provided.
    By default each input is compulsory, in order to make it optional use
    :py:class:`DOptional<deltalanguage.data_types.DOptional>` wrapper.

    .. warning::
        If a node does not have compulsory inputs then it will be evaluated
        continuesly and this can significantly slow down a runtime simulator.
        If node does not have optional inputs consider using
        ``allow_const=True`` to improve performance.

    Parameters
    ----------
    template : NodeTemplate
        Associate this node constructor with this specfied existing node
        template rather than a newly created one.
    name : str
        The name of the node to be made.
    allow_const : bool
        If ``True`` and all inputs are constant then the output of this node is
        calculated only once, cached, and reused at each request.
        This used to reduce the computation load.
    node_key : Optional[str]
        Keyword argument used for providing the node to the block, in case the
        user wants to debug sending & receiving messages in an interactive
        console. Note that this should only be used for debugging; for
        Deltaflow programs in production it should be sufficient to use the
        inputs and return values of a block for communications.
    in_port_size : int
        The maximum size of the node's in ports.
        If 0 then unlimited size.
    latency : Latency
        The estimated latency for running the body.
    lvl : int
        Logging level for the node.


    .. note::
        The function's inputs and outputs have to be typed, otherwise
        the decorator will raise an error.


    Examples
    --------
    The decorated function can be used as a normal python function:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.DeltaBlock()
        ... def foo(a: int) -> int:
        ...     return a + 5

        >>> foo(5)
        10

    The exact same function can be used in the context of
    :py:class:`DeltaGraph`:

    .. code-block:: python

        >>> s = dl.lib.StateSaver(object, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     foo_out = foo(5)
        ...     s.save_and_exit(foo_out) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 10

    Nodes can also have multiple inputs:

    .. code-block:: python

        >>> @dl.DeltaBlock()
        ... def bar(a: int, b: int) -> int:
        ...     return a*b

    Make sure that inputs of such nodes are not mixed up. For this one can
    use keyworded arguments, exactly as in python:

    .. code-block:: python

        >>> with dl.DeltaGraph() as graph:
        ...     s.save_and_exit(bar(a=9, b=9)) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 81

    For multiple outputs please refer to examples for
    :py:func:`make_forked_return<deltalanguage.data_types.make_forked_return>`.
    """
    def decorator(func):
        my_template = NodeTemplate.merge_deltablock(template,
                                                    func,
                                                    allow_const,
                                                    node_key,
                                                    name or func.__name__,
                                                    in_port_size,
                                                    latency,
                                                    lvl)

        @wraps(func)
        def decorated(*args, **kwargs):
            """If there is currently an active :py:class:`DeltaGraph`,
            return a node constructed using the node factory.

            Parameters
            ----------
            args
                Either arguments to be used to evaluate the function or nodes
                to link to this node.
            kwargs
                Either keyworded arguments to be used to evaluate the function
                or nodes to link to this node.

            Returns
            -------
            the result of function evaluation or the created node
            """
            if DeltaGraph.stack():
                # There is currently an active DeltaGraph so use tempalte
                return my_template.call(*args, **kwargs)
            else:
                # No DeltaGraph active so evaluate function as usual
                return func(*args, **kwargs)
        decorated.template = my_template
        return decorated
    return decorator


def DeltaMethodBlock(
    template: NodeTemplate = None,
    name: str = None,
    node_key: Optional[str] = None,
    in_port_size: int = 0,
    latency: Latency = None,
    lvl: int = logging.ERROR
):
    """Decorator to turn a class method to a block for use in
    :py:class:`DeltaGraph`.

    If called in the context of :py:class:`DeltaGraph` it will
    return a node that can have an internal state stored in the instance of
    the target class. Thus the output determined by not only the inputs but
    also by the internal state that can change.

    .. warning::
        The internal state is a very powerful concept that makes the Deltaflow
        language significantly more expressive, but also can lead to
        non-deterministic results. Please refer to
        `Non-deterministic state history <tutorials/state_history.html>`_
        that cover this in detail.

    The node is evaluated when all compulsory inputs are provided.
    By default each input is compulsory, in order to make it optional use
    :py:class:`DOptional<deltalanguage.data_types.DOptional>` wrapper.

    .. warning::
        If a node does not have compulsory inputs then it will be evaluated
        continuesly and this can significantly slow down a runtime simulator.

    Parameters
    ----------
    name : str
        The name of the node to be made.
    node_key : Optional[str]
        Keyword argument used for providing the node to the block, in case the
        user wants to debug sending & receiving messages in an interactive
        console. Note that this should only be used for debugging; for
        Deltaflow programs in production it should be sufficient to use the
        inputs and return values of a block for communications.
    in_port_size : int
        The maximum size of the node's in ports.
        If 0 then unlimited size.
    latency : Latency
        The estimated latency for running the body.
    lvl : int
        Logging level for the node.


    .. note::
        The method's inputs and outputs have to be typed, otherwise
        the decorator will raise an error.


    Examples
    --------
    The decorated method can be used as a normal python class method:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> class MyClass:
        ...
        ...     def __init__(self, x):
        ...         self.x = x
        ...
        ...     @dl.DeltaMethodBlock()
        ...     def bar(self, a: int) -> int:
        ...         return self.x + a

        >>> my_obj = MyClass(10)
        >>> print(my_obj.bar(5))
        15

    The exact same object can be used in the context of
    :py:class:`DeltaGraph`:

    .. code-block:: python

        >>> s = dl.lib.StateSaver(object, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     s.save_and_exit(my_obj.bar(5)) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 15

    However if the internal state of the object changes (before or during),
    the result will change:

    .. code-block:: python

        >>> my_obj.x = 15

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 20

    Nodes can also have multiple inputs exactly as
    :py:class:`DeltaBlock`.

    In case of multiple outputs please refer to examples for
    :py:func:`make_forked_return<deltalanguage.data_types.make_forked_return>`.
    """

    def decorator(func):
        my_template = NodeTemplate.merge_deltamethod(template,
                                                     func,
                                                     node_key,
                                                     name or func.__name__,
                                                     in_port_size,
                                                     latency,
                                                     lvl)

        @wraps(func)
        def decorated(obj, *args, **kwargs):
            """If there is currently an active DeltaGraph, return a node
            constructed using the node factory.

            Parameters
            ----------
            obj
                Instantiated object of the class a_func is from.
            args
                Either arguments to be used to evaluate the function or nodes
                to link to this node.
            kwargs
                Either kwarguments to be used to evaluate the function or nodes
                to link to this node.

            Returns
            -------
            The result of function evaluation or the created node.
            """
            if DeltaGraph.stack():
                # there is currently an active DeltaGraph so use template
                return my_template.call(obj, *args, **kwargs)
            else:
                # No DeltaGraph active so evaluate function as usual
                return func(obj, *args, **kwargs)
        decorated.template = my_template
        return decorated
    return decorator


def Interactive(
    in_params: Dict[str, Type],
    out_type: Type = Void,
    template: NodeTemplate = None,
    name: str = None,
    in_port_size: int = 0,
    latency: Latency = None,
    lvl: int = logging.ERROR
) -> Callable[..., InteractiveBodyTemplate]:
    """Decorator to turn a function of a special type (see examples)
    to an interactive node for use in :py:class:`DeltaGraph`.

    The function used with this decorator is not meant to be used on its own
    (i.e. outside of the graph) in contrast with
    :py:class:`DeltaBlock` and :py:class:`DeltaMethodBlock`.

    The types of inputs and outputs have to be defined using ``in_params`` and
    ``out_type``. By default each input is compulsory, in order to make it
    optional use :py:class:`DOptional<deltalanguage.data_types.DOptional>`
    wrapper.

    Unlike nodes created via :py:class:`DeltaBlock` and
    :py:class:`DeltaMethodBlock`, the interactive node is evaluated only once.
    Thus it's in the user's responsibility to specify when and which
    inputs and outputs are used.
    This is done via ``send`` and ``receive`` methods and shown in examples
    below.

    An interactive node's inputs/outputs can be received/sent simultaneously
    or partially, which gives a lot of flexibility in comparison with
    :py:class:`DeltaBlock` and :py:class:`DeltaMethodBlock`.
    For instance, if a compulsory input is expected then by calling ``receive``
    on this input the node execution will be frozen until the input is
    provided.
    On the other hand, calling ``receive`` for an optional input will not
    freeze the node and simply use ``None`` if no input is provided at
    this instance of time.

    An interactive node should use at least one input or output to communicate
    with other nodes, otherwise runtime simulators, e.g.
    :py:class:`DeltaPySimulator<deltalanguage.runtime.DeltaPySimulator>`,
    might be significantly slowed down by their schedulers.

    Parameters
    ----------
    in_params : Dict[str, Type]
        The types of the in parameters to the node to be made.
    out_type : Type
        The type of the output for the node to be made.
    name : str
        The name of the node to be made.
    in_port_size : int
        The maximum size of the node's in ports.
        If 0 then unlimited size.
    lvl : int
        Logging level for the node.

    Examples
    --------
    Let's look at the syntax first.
    Here is an example where all inputs are received simultaneously several
    times, before the output is sent off once, then the node becomes idle:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.Interactive(in_params={"a": int, "b": int}, out_type=int)
        ... def foo(node: PythonNode):
        ...     internal_memory = 0
        ...
        ...     for i in range(10):
        ...         inputs = node.receive()
        ...         internal_memory += inputs["a"] * inputs["b"]
        ...
        ...     node.send(internal_memory)

    .. note::
        :py:class:`PythonNode` is used to help with syntax highlight.
        For instance when using a linting tool you can access definitions of
        `send` and `receive` methods and see their parameters.

    To use this node in the context of :py:class:`DeltaGraph` a special
    method ``node`` should be used:

    .. code-block:: python

        >>> s = dl.lib.StateSaver(int, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     s.save_and_exit(foo.call(a=4, b=5)) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 200

    Here is the case when inputs are read independently:

    .. code-block:: python

        >>> @dl.Interactive(in_params={"a": int, "b": int}, out_type=int)
        ... def bar(node: PythonNode):
        ...     internal_memory = 0
        ...
        ...     for i in range(10):
        ...         if i%2 == 0:
        ...             internal_memory += node.receive("a")
        ...         else:
        ...             internal_memory -= node.receive("b")
        ...
        ...     node.send(internal_memory)

        >>> with dl.DeltaGraph() as graph:
        ...     s.save_and_exit(bar.call(a=4, b=5)) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving -5

    In case of multiple outputs make sure to call
    :py:func:`make_forked_return<deltalanguage.data_types.make_forked_return>`
    beforehand, then the outputs can be send out in any configuration:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> TwoIntsT, TwoIntsC = dl.make_forked_return({'x': int, 'y': int})

        >>> @dl.Interactive(in_params={"a": int}, out_type=TwoIntsT)
        ... def baz(node: PythonNode):
        ...     for i in range(10):
        ...         a = node.receive("a")
        ...         if a%3 == 0:
        ...             node.send(TwoIntsC(x=a, y=None))
        ...         elif a%3 == 1:
        ...             node.send(TwoIntsC(x=None, y=a))
        ...         else:
        ...             node.send(TwoIntsC(x=a, y=a))
    """
    if not isinstance(in_params, dict):
        raise TypeError('Please provide types of input parameters')

    def decorator(func: Callable[[PythonNode], None]):
        proc = NodeTemplate.merge_interactive(template,
                                              func,
                                              in_params,
                                              out_type,
                                              name or func.__name__,
                                              in_port_size,
                                              latency,
                                              lvl)
        return proc
    return decorator



from copy import deepcopy
import logging
from queue import Empty, Full, Queue
from typing import Any

from deltalanguage.wiring import OutPort
from deltalanguage.logging import make_logger
from deltalanguage._utils import QueueMessage


class Flusher:
    """Empty class is passed down queues to release a blockage."""
    pass


class DeltaQueue(Queue):
    """Queue class that communicate messages between nodes in
    :py:class:`DeltaPySimulator<deltalanguage.runtime.DeltaPySimulator>`.

    The object connects corresponding I/O ports.

    .. note::
        :py:class:`DeltaPySimulator<deltalanguage.runtime.DeltaPySimulator>`
        performs all required splitting of data on the background
        (if required).

    Parameters
    ----------
    out_port : OutPort
        Output port for which this queue is created.
    maxsize : int
        If maxsize is <= 0, the queue size is infinite.
        The default value 16 is chosen in analogy with ``sc_fifo`` in SystemC,
        this means that the Python GIL will be forced to run another node's
        thread once the queue is full, thus it will prevent blockage.
    queue_interval : float
        When the simulation is called to stop, some of the queues might be
        blocking nodes at ``put`` methods as they are full.
        :py:class:`DeltaPySimulator<deltalanguage.runtime.DeltaPySimulator>`
        interrupts this at this periodicity (in seconds) and checks if
        stopping is needed.

    Attributes
    ----------
    optional : bool
        Identify if the corresponding
        :py:class:`InPort<deltalanguage.wiring.InPort>` of the receiving node
        requires a message from this queue or it's optional.
        If ``False`` and nothing is sent down this queue, the caller will use
        ``Queue.get()`` with block and wait.
    _index : Union[str, None]
        For ``ForkedReturn`` this stores the name of
        :py:class:`OutPort<deltalanguage.wiring.OutPort>` providing
        messages for this queue. Otherwise it's ``None``.
    """

    def __init__(self,
                 out_port: OutPort,
                 maxsize: int = 16,
                 queue_interval: float = 1.0):
        super().__init__(maxsize=maxsize)
        self._src = out_port
        self._log = make_logger(logging.WARNING,
                                f"Queue {out_port.port_name}")
        self._queue_interval = queue_interval
        self.optional = out_port.destination.is_optional

        self._index = None
        if out_port.port_name.n_index is not None:
            self._index = out_port.port_name.n_index

    def _index_and_put(self,
                       item: QueueMessage,
                       block=True,
                       timeout=None) -> QueueMessage:
        """In case of ForkedReturn only the requested indexed element will
        be put to the queue, otherwise the entire item.

        ``None`` is not added to the queue.

        If the queue is full, a Full exception is raised after a timeout so
        that the node pushing to it becomes unblocked to check for an exit
        signal.

        Returns a reference to the object added to the queue, which is used
        by ConstQueue for caching.
        """
        if not isinstance(item, QueueMessage):
            raise TypeError("Only QueueMessage objects can be put on queues")

        if self._index is not None:
            msg = item.msg
            to_put_msg = msg[msg._fields.index(self._index)]
            to_put = QueueMessage(msg=to_put_msg, clk=item.clk)
        else:
            to_put = item

        if to_put.msg is not None:
            if timeout is None:
                Queue.put(self, to_put, block, timeout=self._queue_interval)
            else:
                Queue.put(self, to_put, block, timeout=timeout)

        return to_put

    def get(self, block=True, timeout=None) -> QueueMessage:
        """If the queue is optional and empty return ``None``,
        otherwise return the item.
        """
        if self.optional and self.empty():
            return QueueMessage(None, clk=0)

        item = Queue.get(self, block=block, timeout=timeout)

        return item

    def put(self, item: QueueMessage, block=True, timeout=None):
        """Add an item to a queue, performing the indexing if it applies.

        .. warning::
            If ``item.msg == None``, it is not added to the queue.
        """
        self._index_and_put(item, block=block, timeout=timeout)

    def flush(self):
        """Unblock any thread waiting for this queue."""
        if self.empty():
            Queue.put(self, QueueMessage(Flusher(), clk=-1))


class ConstQueue(DeltaQueue):
    """An imitation queue created at the output of
    :py:class:`PyConstBody<deltalanguage.wiring.PyConstBody>`.

    A deepcopy of the message from this queue is passed to the receiving node
    and it is _not_ removed from the queue, thus the next time the receiving
    node ask for the input the same message will be produced without
    re-evaluation of the node this message originated from.

    Parameters
    ----------
    out_port : OutPort
        Output port for which this queue is created.


    .. warning::
        A common misconception about this queue is that the message is passed
        to the receiving node only once, this is not true. You should think
        about an instance of this queue as a continuous stream of identical
        messages. For instance, if the receiving node has only one non-optional
        input from this queue it will be passed to that node at every iteration.
    """

    def __init__(self, out_port):
        super().__init__(out_port, maxsize=1)
        self._saved_value = None

    def put(self, item: QueueMessage, block=True, timeout=None):
        """Overwrite ``DeltaQueue.put``."""
        if self._saved_value is None:
            to_put = self._index_and_put(item, block=block, timeout=timeout)

            # same rule as in _index_and_put: only non-None is saved
            if to_put.msg is not None:
                self._saved_value = to_put
        else:
            raise Full("Put to already populated ConstQueue.")

    def get(self) -> QueueMessage:
        """Overwrite ``DeltaQueue.get``, block and timeout are unavailable."""
        if self.empty():
            if self.optional:
                return QueueMessage(None, clk=0)

            raise Empty

        return deepcopy(self._saved_value)

    def flush(self):
        """Overwrite ``DeltaQueue.flush``"""
        if self.empty():
            self._saved_value = QueueMessage(Flusher(), clk=-1)
            Queue.put(self, self._saved_value)



from deltalanguage.data_types import DUInt, DSize
from deltalanguage.wiring import DeltaMethodBlock, NodeTemplate

from ..quantum_simulators import IQuantumSimulator

hal_template = NodeTemplate(
    name="QSim",
    in_params={'hal_command': DUInt(DSize(32))},
    out_type=DUInt(DSize(32))
)

class HardwareAbstractionLayerNode:
    """Encapsulates a node which receives HAL commands and uses them to
    perform operations on a quantum device.

    Parameters
    ----------
    quantum_simulator : IQuantumSimulator
        Object with the IQuantumSimulator interface that accepts commands
        and returns measurement results.
    """

    def __init__(
        self,
        quantum_simulator: IQuantumSimulator
    ):
        self._quantum_simulator = quantum_simulator

    @DeltaMethodBlock(name="accept_command")
    def accept_command(self, hal_command: DUInt(DSize(32))) -> DUInt(DSize(32)):
        """Interface for ``quantum_simulator.accept_command`` that is used
        to create a graph node.

        Parameters
        ----------
        command : DUInt(DSize(32))
            The HAL command to deconstruct and use to perform actions.

        Returns
        -------
        DUInt(DSize(32))
            Result of a measurement command.
        """
        result = self._quantum_simulator.accept_command(hal_command)

        return result



"""Tests for graph visualisation.
Mainly testing that NetworkX returns the correct form of graph.
Done by comparing nodes and edges. Note that node numbers have
to be stripped, due to the numbers varying depending on how
many graphs are created.
"""

from collections import Counter
import unittest

from deltalanguage.data_types import Void, DOptional
from deltalanguage.runtime import DeltaRuntimeExit
from deltalanguage.wiring import (DeltaBlock, DeltaGraph,
                                  placeholder_node_factory)


@DeltaBlock()
def add(a: int, b: DOptional(int)) -> int:
    if b is None:
        return a
    else:
        return a+b


@DeltaBlock()
def print_then_exit(n: int) -> Void:
    print(n)
    raise DeltaRuntimeExit


@DeltaBlock()
def print_until_10(n: int) -> int:
    print(n)
    if n >= 10:
        raise DeltaRuntimeExit
    else:
        return n


def node_cnt(nodes):
    """Constructs Counter of nodes and strips number from node names."""
    return Counter(["_".join(node.split("_")[:-1]) for node in list(nodes)])


def edge_cnt(edges):
    """Constructs Counter of edges and strips number from node names."""
    return Counter([("_".join(edge[0].split("_")[:-1]),
                     "_".join(edge[1].split("_")[:-1])) for edge in list(edges)])


class SimpleGraph(unittest.TestCase):
    """Testing for a simple graph with four nodes."""

    def test_simple_networkx_graph(self):
        with DeltaGraph() as simple_graph:
            n = add(a=3, b=4)
            print_then_exit(n=n)
        networkx_graph = simple_graph.get_networkx_graph()
        self.assertEqual(node_cnt(networkx_graph.nodes()),
                         Counter({'node': 2, 'add': 1, 'print_then_exit': 1}))
        self.assertEqual(edge_cnt(networkx_graph.edges()),
                         Counter({('node', 'add'): 2,
                                  ('add', 'print_then_exit'): 1}))


class ComplexGraph(unittest.TestCase):
    """Testing for a graph with a cycle."""

    def test_complex_networkx_graph(self):
        with DeltaGraph() as complex_graph:
            ph = placeholder_node_factory()
            n = add(a=1, b=ph)
            print_node = print_until_10(n=n)
            ph.specify_by_node(print_node)
        networkx_graph = complex_graph.get_networkx_graph()
        self.assertEqual(node_cnt(networkx_graph.nodes()),
                         Counter({'node': 1, 'add': 1, 'print_until_10': 1}))
        self.assertEqual(edge_cnt(networkx_graph.edges()),
                         Counter({('node', 'add'): 1,
                                  ('add', 'print_until_10'): 1,
                                  ('print_until_10', 'add'): 1}))


if __name__ == "__main__":
    unittest.main()
